"""
Curriculum Test Suite
"""

import unittest

from pydantic import BaseModel, Field

from aind_behavior_curriculum import Metrics, TaskParameters
from aind_behavior_curriculum.curriculum import (
    _NonDeserializableCallable,
    _Rule,
    is_non_deserializable_callable,
    try_materialize_non_deserializable_callable_error,
)


def rule_update(metrics: Metrics, params: TaskParameters) -> TaskParameters:
    return params


def simple_rule(m, p):
    return m + p


def duck_type_rule_update(m, p):
    return p


def not_a_rule_update(metrics: int, params: BaseModel) -> BaseModel:
    return params


class RuleTests(unittest.TestCase):
    def setUp(self):
        class CustomRule(_Rule[[Metrics, TaskParameters], TaskParameters]):
            pass

        class Container(BaseModel):
            this_new_rule: CustomRule = Field(default=CustomRule(rule_update))

        self.container = Container
        self.custom_rule = CustomRule

    def test_rule_instantiation(self):
        self.container(this_new_rule=self.custom_rule(rule_update))

    def test_duck_type_rule_instantiation(self):
        self.container(this_new_rule=self.custom_rule(duck_type_rule_update))

    def test_not_a_rule_instantiation(self):
        with self.assertRaises(TypeError):
            self.container(this_new_rule=self.custom_rule(not_a_rule_update))

    def test_can_serialize_rule(self):
        container = self.container(this_new_rule=self.custom_rule(rule_update))
        container.model_dump()
        container.model_dump_json()

    def test_can_deserialize_rule(self):
        container = self.container(this_new_rule=self.custom_rule(rule_update))
        dump = container.model_dump()
        self.assertEqual(container, container.model_validate(dump))
        json_dump = container.model_dump_json()
        self.assertEqual(container, container.model_validate_json(json_dump))

    def test_can_deserialize_rule_callable(self):
        container = self.container(this_new_rule=self.custom_rule(simple_rule))
        self.assertEqual(container.this_new_rule.invoke(0, 0), 0)

        dump = container.model_dump()
        deser_dump = container.model_validate(dump)
        self.assertEqual(container, deser_dump)
        self.assertEqual(deser_dump.this_new_rule.invoke(0, 0), 0)

        json_dump = container.model_dump_json()
        deser_json = container.model_validate_json(json_dump)
        self.assertEqual(container, deser_json)
        self.assertEqual(deser_dump.this_new_rule.invoke(0, 0), 0)

    def test_is_non_deserializable_callable(self):
        callable_ref = _NonDeserializableCallable("test", Exception("test"))
        with self.assertRaises(RuntimeError):
            callable_ref()
        self.assertFalse(is_non_deserializable_callable(self.custom_rule(rule_update)))
        self.assertTrue(is_non_deserializable_callable(callable_ref))

        self.assertIsInstance(
            try_materialize_non_deserializable_callable_error(callable_ref),
            Exception,
        )
        self.assertIsNone(try_materialize_non_deserializable_callable_error(self.custom_rule(rule_update)))

    def test_can_deserialize_rule_reference(self):
        def rule_update_to_be_deleted(metrics: Metrics, params: TaskParameters) -> TaskParameters:
            return params

        def back_rule_update(metrics: Metrics, params: TaskParameters) -> TaskParameters:
            return params

        container = self.container(this_new_rule=self.custom_rule(rule_update_to_be_deleted))
        dump = container.model_dump()
        json_dump = container.model_dump_json()

        del rule_update_to_be_deleted

        deser = container.model_validate(dump)
        deser_json = container.model_validate_json(json_dump)

        self.assertTrue(is_non_deserializable_callable(deser.this_new_rule.callable))
        self.assertTrue(is_non_deserializable_callable(deser_json.this_new_rule.callable))

    def test_rule_from_callable(self):
        container = self.container(this_new_rule=rule_update)
        dump = container.model_dump()
        json_dump = container.model_dump_json()

        deser = container.model_validate(dump)
        deser_json = container.model_validate_json(json_dump)

        self.assertEqual(container, deser)
        self.assertEqual(container, deser_json)

    def test_rule_from_not_callable(self):
        with self.assertRaises(TypeError):
            _ = self.container(this_new_rule=not_a_rule_update)


if __name__ == "__main__":
    unittest.main()



"""
Example of Curriculum creation
"""

import json

from pydantic import Field

from aind_behavior_curriculum import (
    GRADUATED,
    Graduated,
    Metrics,
    Policy,
    PolicyTransition,
    Stage,
    StageTransition,
    Task,
    TaskParameters,
    create_curriculum,
    create_task,
)
from aind_behavior_curriculum.curriculum_utils import (
    export_diagram,
    export_json,
)


def init_stage_rule(metrics: Metrics, task: Task) -> Task:
    """
    Trivially pass the default
    """
    return task


INIT_STAGE = Policy(init_stage_rule)


# --- TASKS ---
class TaskAParameters(TaskParameters):
    field_a: int = Field(default=0, validate_default=True)


TaskA = create_task(name="Task A", task_parameters=TaskAParameters)


class TaskBParameters(TaskParameters):
    field_b: float = Field(default=0.0)


TaskB = create_task(name="Task B", task_parameters=TaskBParameters)


# --- METRICS ---
class ExampleMetrics(Metrics):
    """
    Totally made up values we will edit ourselves to simulate mouse training.
    Each theta value is reserved for a test case.
    """

    theta_1: int = Field(default=0)
    theta_2: int = Field(default=0)
    theta_3: int = Field(default=0)


# --- POLICIES ---
def stageA_policyA_rule(metrics: ExampleMetrics, task: TaskA) -> TaskA:
    task = task.model_copy(deep=True)
    task.task_parameters.field_a = 8
    return task


stageA_policyA = Policy(stageA_policyA_rule)


def stageA_policyB_rule(metrics: ExampleMetrics, task: TaskA) -> TaskA:
    task = task.model_copy(deep=True)
    task.task_parameters.field_a = 16
    return task


stageA_policyB = Policy(stageA_policyB_rule)


def stageB_policyA_rule(metrics: ExampleMetrics, task: TaskB) -> TaskB:
    task = task.model_copy(deep=True)
    task.task_parameters.field_b = 8
    return task


stageB_policyA = Policy(stageB_policyA_rule)


def stageB_policyB_rule(metrics: ExampleMetrics, task: TaskB) -> TaskB:
    task = task.model_copy(deep=True)
    task.task_parameters.field_b = 16
    return task


stageB_policyB = Policy(stageB_policyB_rule)


# --- POLICY TRANSITIONS ---
def t1_5_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_1 > 5


t1_5 = PolicyTransition(t1_5_rule)


def t1_10_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_1 > 10


t1_10 = PolicyTransition(t1_10_rule)


def t3_5_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_3 > 5


t3_5 = PolicyTransition(t3_5_rule)


def t3_10_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_3 > 10


t3_10 = PolicyTransition(t3_10_rule)


# --- STAGE TRANSITIONS ---
def t2_5_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_2 > 5


t2_5 = StageTransition(t2_5_rule)


def t2_10_rule(metrics: ExampleMetrics) -> bool:
    return metrics.theta_2 > 10


t2_10 = StageTransition(t2_10_rule)


# --- CURRICULUM ---


MyCurriculum = create_curriculum(name="My Curriculum", version="0.1.0", tasks=(TaskA, TaskB, Graduated))


def construct_curriculum():
    """
    Useful for testing.
    """

    # Init Stages
    taskA = TaskA(task_parameters=TaskAParameters())
    taskB = TaskB(task_parameters=TaskBParameters())
    stageA = Stage(name="StageA", task=taskA)
    stageB = Stage(name="StageB", task=taskB)

    stageA.add_policy_transition(INIT_STAGE, stageA_policyB, t1_10)
    stageA.add_policy_transition(INIT_STAGE, stageA_policyA, t1_5)
    stageA.add_policy_transition(stageA_policyA, stageA_policyB, t1_10)
    stageA.set_start_policies(INIT_STAGE)

    stageB.add_policy_transition(INIT_STAGE, stageB_policyB, t3_10)
    stageB.add_policy_transition(INIT_STAGE, stageB_policyA, t3_5)
    stageB.add_policy_transition(stageB_policyA, stageB_policyB, t3_10)
    stageB.set_start_policies(INIT_STAGE)

    # Construct the Curriculum
    ex_curr = MyCurriculum(name="My Curriculum")
    ex_curr.add_stage_transition(stageA, GRADUATED, t2_10)
    ex_curr.add_stage_transition(stageA, stageB, t2_5)
    ex_curr.add_stage_transition(stageB, GRADUATED, t2_10)

    return ex_curr


if __name__ == "__main__":
    ex_curr = construct_curriculum()
    name = "curriculum"
    with open(
        f"./examples/example_project/assets/{name}_schema.json",
        "w+",
        encoding="utf-8",
    ) as f:
        f.write(json.dumps(ex_curr.model_json_schema(), indent=4))
    export_json(ex_curr, path=f"./examples/example_project/assets/{name}.json")
    _ = export_diagram(ex_curr, f"./examples/example_project/assets/{name}.svg")



"""
Base Behavior Models
"""

from string import capwords
from typing import Annotated, Generic, Literal, Optional, Type, TypeVar

from pydantic import Field, SerializeAsAny, create_model

from aind_behavior_curriculum.base import (
    AindBehaviorModel,
    AindBehaviorModelExtra,
)

SEMVER_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"


TTaskParameters = TypeVar("TTaskParameters", bound=AindBehaviorModelExtra)

TaskParameters = AindBehaviorModelExtra


class Task(AindBehaviorModel, Generic[TTaskParameters]):
    """
    Base class for all Tasks. A new Task can be created by:
        - Subclassing: class NewTask(TaskParameters)
        - Using the factory method: create_task(...)
        - Passing a type to the generic base: NewTask = Task[MyParameters]
    Holds Task metadata and parameters.
    """

    name: str = Field(..., description="Name of the task.", frozen=True)
    description: str = Field(default="", description="Description of the task.")
    task_parameters: SerializeAsAny[TTaskParameters]
    version: Optional[str] = Field(
        default=None,
        pattern=SEMVER_REGEX,
        description="task schema version",
        frozen=True,
    )
    stage_name: Optional[str] = Field(
        default=None,
        description="Optional stage name the `Task` object instance represents.",
    )


TTask = TypeVar("TTask", bound="Task")


def create_task(
    *,
    name: str,
    task_parameters: Type[TTaskParameters],
    version: Optional[str] = None,
    description: str = "",
) -> Type[Task[TTaskParameters]]:
    """
    Factory method for creating a Task object.

    Args:
        name: Name of the task.
        task_parameters: Task parameters.
        version: Task schema version.
        description: Task description.
        stage_name: Optional stage name the `Task` object instance represents.

    Returns:
        Task: Task object instance.
    """

    def _snake_to_pascal(v: str) -> str:
        """Converts a string from snake_case to PascalCase"""
        return "".join(map(capwords, v.split("_")))

    return create_model(
        _snake_to_pascal(name),
        __base__=Task[task_parameters],
        name=Annotated[
            Literal[name],
            Field(default=name, frozen=True, validate_default=True),
        ],
        version=Annotated[
            Literal[version] if version else Optional[str],
            Field(
                default=version,
                frozen=True,
                pattern=SEMVER_REGEX,
                validate_default=True,
            ),
        ],
        description=Annotated[
            str,
            Field(default=description, frozen=True, validate_default=True),
        ],
    )



"""
Core Trainer primitive.
"""

from abc import abstractmethod
from collections.abc import Iterable
from functools import reduce
from typing import Annotated, Generic, List, Optional, Self, Type, TypeVar

from pydantic import Field, create_model

from aind_behavior_curriculum.base import AindBehaviorModel
from aind_behavior_curriculum.curriculum import (
    Curriculum,
    Metrics,
    Policy,
    Stage,
    StageGraph,
    Task,
    make_task_discriminator,
)

TCurriculum = TypeVar("TCurriculum", bound=Curriculum)
TMetrics = TypeVar("TMetrics", bound=Metrics)
TTask = TypeVar("TTask", bound=Task)


class TrainerState(AindBehaviorModel, Generic[TCurriculum]):
    """
    Trainer State.
    Pydantic model for de/serialization.
    """

    curriculum: Optional[TCurriculum] = Field(
        ...,
        validate_default=True,
        description="The curriculum used by the trainer",
    )
    # Note: This will deserialize to a base Stage object.
    # Should users require the subclass, they are incentivized to use
    # the Trainer.create_trainer_state property instead
    stage: Optional[Stage] = Field(
        ...,
        validate_default=True,
        description="The output suggestion of the curriculum",
    )
    is_on_curriculum: bool = Field(
        default=True,
        validate_default=True,
        description="Was the output suggestion generated as part of the curriculum?",
    )
    active_policies: Optional[List[Policy]] = Field(
        default=[],
        validate_default=True,
        description="The active policies for the current stage",
    )

    @classmethod
    def default(cls) -> Self:
        """
        Class method to create a default instance of the class.

        Returns:
            Self: An instance of the class with default parameters.
        """
        return cls(
            curriculum=None,
            stage=None,
            is_on_curriculum=False,
            active_policies=None,
        )

    def __eq__(self, other: object) -> bool:
        """
        TrainerState Equality
        """
        if not isinstance(other, TrainerState):
            raise NotImplementedError("Equality comparison only implemented for _TrainerState objects.")

        # Compare 'stage' and 'is_on_curriculum' attributes
        if self.stage != other.stage or self.is_on_curriculum != other.is_on_curriculum:
            return False

        # Compare active_policies using set equality
        if self.active_policies is None and other.active_policies is None:
            return True
        if self.active_policies is None or other.active_policies is None:
            return False

        # Extract Rule callable which is hashable for set equality
        self_rules = [p for p in self.active_policies]
        other_rules = [p for p in other.active_policies]

        return set(self_rules) == set(other_rules)


class Trainer(Generic[TCurriculum]):
    """
    Trainer class for managing and evaluating curriculum stages and policy transitions,
    and updating the task parameters based on the active policies and provided metrics.
    The entry point is the "evaluate" method.
    Attributes:
        curriculum (TCurriculum): The curriculum used by the trainer.
    Methods:
        __init__(self, curriculum: TCurriculum):
        curriculum(self) -> TCurriculum:
        _evaluate_stage_transition(curriculum: Curriculum, current_stage: Stage, metrics: TMetrics) -> Optional[Stage]:
        _evaluate_policy_transitions(cls, current_stage: Stage, active_policies: Iterable[Policy], metrics: TMetrics) -> List[Policy]:
            Evaluates policy transitions for the given current stage and currently active policies, based on the provided metrics.
        evaluate(self, trainer_state: TrainerState, metrics: TMetrics) -> TrainerState:
        get_net_parameter_update(stage_parameters: TaskParameters, stage_policies: Iterable[Policy], curr_metrics: Metrics) -> TaskParameters:
            Aggregates parameter updates of input stage_policies given current stage_parameters and current metrics.
        _get_unique_policies(policies: List[Policy]) -> List[Policy]:
            Filters unique policies based on their rule functions and reassembles the Policy objects.
    """

    def __init__(self, curriculum: TCurriculum):
        """
        Initializes the Trainer with the given curriculum.
        Args:
            curriculum (TCurriculum): The curriculum to be used by the trainer.
        """

        self._curriculum = curriculum
        self._trainer_state_factory = self._construct_trainer_state_type_from_curriculum(curriculum)

    @property
    def curriculum(self) -> TCurriculum:
        """
        Property that returns the current curriculum.

        Returns:
            TCurriculum: The current curriculum instance.
        """
        return self._curriculum

    def create_enrollment(self) -> TrainerState[TCurriculum]:
        """Creates a new TrainerState for a subject enrolling in the curriculum.
        The initial stage is determined by the curriculum's graph, by
        selecting the stage with the lowest index.

        Returns:
            TrainerState: A new instance of the type-aware TrainerState class,
            initialized to the first stage of the curriculum.
        """

        stages: StageGraph[Metrics, Task] = self.curriculum.graph
        if len(stages.see_nodes()) == 0:
            raise ValueError("Curriculum has no stages.")
        min_index = min(stages.nodes.keys())
        init_stage = stages.nodes[min_index]
        return self.create_trainer_state(
            is_on_curriculum=True,
            active_policies=init_stage.start_policies,
            stage=init_stage,
        )

    @property
    def trainer_state_model(self) -> Type[TrainerState[TCurriculum]]:
        """
        Property that returns a type-aware TrainerState class.

        Returns:
            Type[TrainerState]: type-aware TrainerState type.
        """
        return self._trainer_state_factory

    def create_trainer_state(
        self,
        *,
        stage: Optional[Stage],
        is_on_curriculum: bool = True,
        active_policies: Optional[Iterable[Policy]] = None,
    ) -> TrainerState[TCurriculum]:
        """
        Creates a new instance of the type-aware TrainerState class.

        Returns:
            TrainerState: A new instance of the type-aware TrainerState class.
        """
        return self._trainer_state_factory(
            curriculum=self.curriculum,
            stage=stage,
            is_on_curriculum=is_on_curriculum,
            active_policies=list(active_policies) if active_policies else None,
        )

    @staticmethod
    def _construct_trainer_state_type_from_curriculum(
        curriculum: TCurriculum,
    ) -> Type[TrainerState[TCurriculum]]:
        """Constructs a task-type-aware TrainerState"""
        _union_type = make_task_discriminator(curriculum._known_tasks)

        return create_model(
            f"{curriculum.name}TrainerState",
            __base__=TrainerState[type(curriculum)],
            stage=Annotated[
                Optional[Stage[Metrics, _union_type]],
                Field(frozen=True, validate_default=True),
            ],
        )

    @staticmethod
    def _evaluate_stage_transition(curriculum: TCurriculum, current_stage: Stage, metrics: TMetrics) -> Optional[Stage]:
        """
        Evaluates whether a transition to a new stage is needed based on the given metrics.

        Args:
            curriculum (Curriculum): The curriculum containing the stage transitions.
            current_stage (Stage): The current stage of the curriculum.
            metrics (TMetrics): The metrics used to evaluate the stage transition.

        Returns:
            Optional[Stage]: The new stage if a transition is made, otherwise None.
        """
        updated_stage: Optional[Stage] = None
        # This line binds the Stage object to the curriculum.
        # TODO may be worth finding a better way to hash the stage object.
        stage_transitions = curriculum.see_stage_transitions(current_stage)
        for stage_eval, dest_stage in stage_transitions:
            # On the first (and only first) true evaluation we transition.
            if stage_eval.invoke(metrics):  # type: ignore
                updated_stage = dest_stage
                break
        return updated_stage

    @classmethod
    def _evaluate_policy_transitions(
        cls,
        current_stage: Stage[TMetrics, TTask],
        active_policies: Iterable[Policy[TMetrics, TTask]],
        metrics: TMetrics,
    ) -> List[Policy[TMetrics, TTask]]:
        """
        Evaluates policy transitions, for the given current stage and currently active policies, based on the provided metrics.
        Args:
            current_stage (Stage): The current stage in the curriculum.
            active_policies (Iterable[Policy]): Currently active policies.
            metrics (TMetrics): The metrics used to evaluate policy transitions.
        Returns:
            List[Policy]: a list of unique policies that are active after the evaluation.
        """
        # Buffer data structures to store result of active policy transitions.
        dest_policies: list[Policy[TMetrics, TTask]] = []

        for active_policy in active_policies:
            policy_transitions = current_stage.see_policy_transitions(active_policy)

            _has_transitioned = False
            for policy_eval, dest_policy in policy_transitions:
                # On first true evaluation, add to buffers
                # and evaluate next active_policy.
                if policy_eval.invoke(metrics):  # type: ignore
                    dest_policies.append(dest_policy)
                    _has_transitioned = True
                    break  # onto next active policy
            if not _has_transitioned:  # if no policy transition keep the current one
                dest_policies.append(active_policy)

        return cls._get_unique_policies(dest_policies)

    def evaluate(self, trainer_state: TrainerState[TCurriculum], metrics: Metrics) -> TrainerState[TCurriculum]:
        """
        Evaluates the current state of the trainer and updates the stage and policies based on the provided metrics.
        Args:
            trainer_state (TrainerState): The current state of the trainer, including the current stage and active policies.
            metrics (TMetrics): The metrics used to evaluate the current state and determine transitions.
        Returns:
            TrainerState: The updated state of the trainer, including the new stage and active policies.
        Raises:
            ValueError: If the current stage or active policies are not set in the trainer state.
        """
        current_stage = trainer_state.stage
        active_policies: Optional[Iterable[Policy[Metrics, Task]]] = trainer_state.active_policies

        if current_stage is None:
            raise ValueError("No current stage. This likely means subject is off-curriculum.")

        # 1) Evaluate stage transitions
        updated_stage = self._evaluate_stage_transition(self.curriculum, current_stage, metrics)

        # 2) Evaluate policy transitions
        # If we've already transitioned stages, we don't need to check policies.
        if updated_stage is None:
            updated_stage = current_stage

            active_policies = active_policies if active_policies is not None else []

            active_policies = self._evaluate_policy_transitions(current_stage, active_policies, metrics)
            # 3) Bootstrap updated parameters with new policies
            updated_task = self.get_net_parameter_update(updated_stage.get_task(), active_policies, metrics)
            updated_stage.set_task(updated_task)

        # If we've transitioned stages, we keep to default task_parameters,
        # and reset active_policies to the start_policies of the new stage.
        else:
            active_policies = updated_stage.start_policies

        return self._trainer_state_factory(
            curriculum=self.curriculum,
            stage=updated_stage,
            is_on_curriculum=True,
            active_policies=active_policies,
        )

    @staticmethod
    def get_net_parameter_update(
        task: TTask,
        stage_policies: Iterable[Policy[TMetrics, TTask]],
        curr_metrics: TMetrics,
    ) -> TTask:
        """
        Aggregates parameter update of input stage_policies
        given current stage_parameters and current metrics.
        """

        return reduce(
            lambda t, p: Policy.normalize_rule_or_callable(p).invoke(curr_metrics, t),
            stage_policies,
            task,
        )

    @staticmethod
    def _get_unique_policies(policies: List[Policy[TMetrics, TTask]]) -> List[Policy[TMetrics, TTask]]:
        """
        set(policies) is not hashable, although Policy
        only contains a function, which is hashable.
        This utility filters on policy functions and
        reassembles the Policy objects.
        """

        filtered_funcs = list(set(p for p in policies))
        output = [Policy[TMetrics, TTask](f) for f in filtered_funcs]

        return output


class TrainerServer:
    """
    Pulls subject curriculum and history and performs evaluation.

    Intended usage:
    1) Implement abstract methods
    2) Call Trainer.register_subject() x N
    3) Call Trainer.evaluate_subject() or Trainer.override_subject_status() x N
    """

    def __init__(self):
        """
        Trainer manages a list of subjects initialized here.
        NOTE: Within Trainer subclass, please call super().__init__()
        """
        self.subject_ids: List[int] = []

    @abstractmethod
    def load_data(self, subject_id: int) -> tuple[Curriculum, TrainerState, Metrics]:
        """
        User-defined.
        Loads 3 pieces of data in the following format:
        - subject Curriculum
        - subject Trainer State
        - subject Metrics
        """
        raise NotImplementedError

    @abstractmethod
    def write_data(
        self,
        subject_id: int,
        curriculum: Curriculum,
        trainer_state: TrainerState,
    ) -> None:
        """
        User-defined.
        Exports 3 pieces of data to database.
        - subject Id
        - subject Curriculum
        - subject Trainer State

        """
        raise NotImplementedError

    def _update_subject_trainer_state(
        self,
        s_id: int,
        curriculum: Curriculum,
        stage: Optional[Stage[TMetrics, TTask]],
        updated_task: Optional[TTask],
        stage_policies: Optional[Iterable[Policy[TMetrics, TTask]]],
    ) -> None:
        """
        Updates subject history, which involves many steps.
        Stage parameters and policies are expected to be part
        of stage-- not checked here b/c this is a private utility.

        If any of {stage, updated_stage_parameters, stage_policies} are None,
        all of the elements are expected to be None.
        """
        if not (stage is None or updated_task is None or stage_policies is None):
            stage.set_task(updated_task)

        trainer = Trainer(curriculum)
        if stage is None:
            trainer_state = trainer.create_trainer_state(stage=None, is_on_curriculum=False, active_policies=None)
        else:
            trainer_state = trainer.create_trainer_state(
                stage=stage,
                is_on_curriculum=True,
                active_policies=(list(stage_policies) if stage_policies else None),
            )

        self.write_data(s_id, curriculum, trainer_state)

    def register_subject(
        self,
        subject_id: int,
        curriculum: Curriculum[TTask],
        start_stage: Stage[TMetrics, TTask],
        start_policies: Optional[Policy[TMetrics, TTask] | List[Policy[TMetrics, TTask]]] = None,
    ) -> None:
        """
        Adds subject into the Trainer system.
        If start_policies is None,
        registration defaults to the Stage.start_policies.
        """

        curriculum = curriculum.validate_curriculum()

        if start_stage not in curriculum.see_stages():
            raise ValueError("Provided start_stage is not in provided curriculum.")
        if subject_id in self.subject_ids:
            raise ValueError(f"Subject_id {subject_id} is already registered.")

        if start_policies is None:
            start_policies = start_stage.start_policies
        elif isinstance(start_policies, Policy):
            start_policies = [start_policies]

        for s_policy in start_policies:
            if s_policy not in start_stage.see_policies():
                raise ValueError(f"Provided start_policy {s_policy} not in provided start_stage {start_stage.name}.")

        _start_policies = list(start_policies)

        initial_task = Trainer.get_net_parameter_update(
            start_stage.get_task(),
            _start_policies,
            curr_metrics=Metrics(),  # Metrics is empty on registration.
        )
        self._update_subject_trainer_state(
            subject_id,
            curriculum,
            start_stage,
            initial_task,
            _start_policies,
        )

        # Add to trainer's local list!
        self.subject_ids.append(subject_id)

    def evaluate_subjects(self) -> None:
        """
        Calls user-defined functions to automatically update
        subject stage along curriculum.
        The time-step between evaluate_subject calls is flexible--
        this function will skip subject to the latest stage/policy
        they are applicable for.

        Evaluation checks for stage transitions before policy transitions.

        If subject does not satisfy any transition criteria,
        this method creates a duplicate current (stage, policy) entry
        in stage history.
        """

        for s_id in self.subject_ids:
            curriculum, trainer_state, curr_metrics = self.load_data(s_id)
            trainer = Trainer(curriculum)

            if trainer_state.stage is not None:
                updated_trainer_state = trainer.evaluate(trainer_state, curr_metrics)
                if updated_trainer_state.stage is None:
                    raise ValueError("Trainer.evaluate() returned None stage. This should not happen.")
                updated_task = updated_trainer_state.stage.get_task()
            else:
                updated_trainer_state = trainer.create_trainer_state(
                    stage=None,
                    is_on_curriculum=False,
                    active_policies=trainer_state.active_policies,
                )  # Not sure if this is correct, but a user may want to keep track of the policies that were active when the subject was off-curriculum.
                updated_task = None

            self._update_subject_trainer_state(
                s_id,
                curriculum,
                updated_trainer_state.stage,
                updated_task,
                updated_trainer_state.active_policies,
            )

    def override_subject_status(
        self,
        s_id: int,
        override_stage: Stage,
        override_policies: Policy | list[Policy],
    ) -> None:
        """
        Override subject (stage, policies) independent of evaluation.
        Stage and Policy objects may be accessed by calling
        Trainer.load_data and looking inside of the returned Curriculum.

        (Soft Rejection-- send mouse to Stage/Policy w/in Curriculum)
        """
        if s_id not in self.subject_ids:
            raise ValueError(f"subject id {s_id} not in self.subject_ids.")

        curriculum, _, curr_metrics = self.load_data(s_id)

        if override_stage not in curriculum.see_stages():
            raise ValueError(
                f"Override stage {override_stage.name} not in curriculum.\
                curriculum stages for subject id {s_id}."
            )

        if not isinstance(override_policies, Iterable):
            override_policies = [override_policies]
        for o_policy in override_policies:
            o_policy = Policy.normalize_rule_or_callable(o_policy)
            if o_policy not in override_stage.see_policies():
                raise ValueError(
                    f"Override policy {o_policy} not in \
                    given override stage {override_stage.name}."
                )

        # Update Stage parameters according to override policies
        updated_params = Trainer(curriculum).get_net_parameter_update(
            override_stage.get_task(),
            override_policies,
            curr_metrics,
        )
        self._update_subject_trainer_state(
            s_id,
            curriculum,
            override_stage,
            updated_params,
            list(override_policies),
        )

    def eject_subject(self, s_id: int) -> None:
        """
        Send mouse off curriculum.
        Only way to get mouse back into system is
        with Trainer.override_subject_status(...)
        """

        curriculum, _, _ = self.load_data(s_id)

        self._update_subject_trainer_state(
            s_id,
            curriculum,
            stage=None,
            updated_task=None,
            stage_policies=None,
        )



"""
Example of Curriculum creation
"""

import json
from typing import Literal

from pydantic import Field

from aind_behavior_curriculum import (
    Curriculum,
    Metrics,
    Policy,
    PolicyTransition,
    Stage,
    StageGraph,
    StageTransition,
    Task,
    TaskParameters,
    make_task_discriminator,
)
from aind_behavior_curriculum.curriculum_utils import (
    export_diagram,
    export_json,
)


# --- TASKS ---
class DummyParameters(TaskParameters):
    field_1: int = Field(default=0, validate_default=True)
    field_2: int = Field(default=0, validate_default=True)


class DummyTask(Task):
    name: Literal["DummyTask"] = "DummyTask"
    task_parameters: DummyParameters = Field(..., description="Fill w/ Parameter Defaults", validate_default=True)


# --- METRICS ---
class ExampleMetrics2(Metrics):
    """
    Totally made up values we will edit ourselves to simulate mouse training.
    Each theta value is reserved for a test case.
    """

    m1: int = Field(default=0)
    m2: int = Field(default=0)


# --- POLICIES ---
# Policies 1-3 do the same thing
# Policies 4-6 do the same thing
def policy_1_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_1 += 5
    return task


policy_1 = Policy(policy_1_rule)


def policy_2_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_1 += 5
    return task


policy_2 = Policy(policy_2_rule)


def policy_3_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_1 += 5
    return task


policy_3 = Policy(policy_3_rule)


def policy_4_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_2 += 5
    return task


policy_4 = Policy(policy_4_rule)


def policy_5_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_2 += 5
    return task


policy_5 = Policy(policy_5_rule)


def policy_6_rule(metrics: ExampleMetrics2, task: DummyTask) -> DummyTask:
    task = task.model_copy(deep=True)
    task.task_parameters.field_2 += 5
    return task


policy_6 = Policy(policy_6_rule)


# --- POLICY/STAGE TRANSITIONS ---
def m1_rule(metrics: ExampleMetrics2) -> bool:
    return metrics.m1 > 0


m1_policy_transition = PolicyTransition(m1_rule)
m1_stage_transition = StageTransition(m1_rule)


def m2_rule(metrics: ExampleMetrics2) -> bool:
    return metrics.m2 > 0


m2_policy_transition = PolicyTransition(m2_rule)
m2_stage_transition = StageTransition(m2_rule)


# --- CURRICULUM ---
Tasks = make_task_discriminator(tasks=[DummyTask])


class MyCurriculum(Curriculum):
    name: Literal["My Curriculum"] = "My Curriculum"
    graph: StageGraph[Metrics, Tasks] = Field(default=StageGraph[Metrics, Tasks]())  # type: ignore


def construct_track_curriculum() -> MyCurriculum:
    dummy_task = DummyTask(task_parameters=DummyParameters())
    test_stage = Stage(name="Track Stage", task=dummy_task)

    test_stage.add_policy_transition(policy_1, policy_2, m1_policy_transition)
    test_stage.add_policy_transition(policy_2, policy_3, m1_policy_transition)
    test_stage.add_policy_transition(policy_4, policy_5, m2_policy_transition)
    test_stage.add_policy_transition(policy_5, policy_6, m2_policy_transition)
    test_stage.set_start_policies([policy_1, policy_4])

    test_curr = MyCurriculum(name="My Curriculum")
    test_curr.add_stage(test_stage)

    return test_curr


def construct_tree_curriculum() -> MyCurriculum:
    dummy_task = DummyTask(task_parameters=DummyParameters())
    test_stage = Stage(name="Tree Stage", task=dummy_task)

    test_stage.add_policy_transition(policy_1, policy_4, m1_policy_transition)
    test_stage.add_policy_transition(policy_2, policy_4, m1_policy_transition)
    test_stage.add_policy_transition(policy_2, policy_5, m1_policy_transition)
    test_stage.add_policy_transition(policy_3, policy_5, m1_policy_transition)
    test_stage.add_policy_transition(policy_4, policy_6, m2_policy_transition)
    test_stage.add_policy_transition(policy_5, policy_6, m2_policy_transition)
    test_stage.set_start_policies([policy_1, policy_2, policy_3])

    test_curr = MyCurriculum(name="My Curriculum")
    test_curr.add_stage(test_stage)

    return test_curr


def construct_policy_triangle_curriculum() -> MyCurriculum:
    dummy_task = DummyTask(task_parameters=DummyParameters())
    test_stage = Stage(name="Triangle Stage", task=dummy_task)

    # Counter-clockwise, higher priority
    test_stage.add_policy_transition(policy_1, policy_2, m1_policy_transition)
    test_stage.add_policy_transition(policy_2, policy_3, m1_policy_transition)
    test_stage.add_policy_transition(policy_3, policy_1, m1_policy_transition)

    # Clockwise
    test_stage.add_policy_transition(policy_1, policy_3, m2_policy_transition)
    test_stage.add_policy_transition(policy_3, policy_2, m2_policy_transition)
    test_stage.add_policy_transition(policy_2, policy_1, m2_policy_transition)

    test_stage.set_start_policies([policy_1])

    test_curr = MyCurriculum(name="My Curriculum")
    test_curr.add_stage(test_stage)

    return test_curr


def construct_stage_triangle_curriculum() -> MyCurriculum:
    dummy_task = DummyTask(task_parameters=DummyParameters())

    test_stage_1 = Stage(name="Stage 1", task=dummy_task)
    test_stage_2 = Stage(name="Stage 2", task=dummy_task)
    test_stage_3 = Stage(name="Stage 3", task=dummy_task)

    test_curr = MyCurriculum(name="My Curriculum")
    # Counter-clockwise, higher priority
    test_curr.add_stage_transition(test_stage_1, test_stage_2, m1_stage_transition)
    test_curr.add_stage_transition(test_stage_2, test_stage_3, m1_stage_transition)
    test_curr.add_stage_transition(test_stage_3, test_stage_1, m1_stage_transition)

    # Clockwise
    test_curr.add_stage_transition(test_stage_1, test_stage_3, m2_stage_transition)
    test_curr.add_stage_transition(test_stage_3, test_stage_2, m2_stage_transition)
    test_curr.add_stage_transition(test_stage_2, test_stage_1, m2_stage_transition)

    return test_curr


if __name__ == "__main__":
    for h in [
        construct_track_curriculum,
        construct_tree_curriculum,
        construct_policy_triangle_curriculum,
        construct_stage_triangle_curriculum,
    ]:
        curriculum = h()
        name = h.__name__.removeprefix("construct_")
        with open(
            f"./examples/example_project_2/assets/{name}_schema.json",
            "w+",
            encoding="utf-8",
        ) as f:
            f.write(json.dumps(curriculum.model_json_schema(), indent=4))
        export_json(curriculum, path=f"./examples/example_project_2/assets/{name}.json")
        _ = export_diagram(curriculum, f"./examples/example_project_2/assets/{name}.svg")



"""
Core Stage and Curriculum Primitives.
"""

import importlib
import inspect
import warnings
from types import EllipsisType
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    ParamSpec,
    Self,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pydantic import Field, GetJsonSchemaHandler, ValidationError, create_model
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema
from typing_extensions import TypeAliasType, cast, deprecated, get_args, get_origin

from aind_behavior_curriculum.base import (
    AindBehaviorModel,
    AindBehaviorModelExtra,
)
from aind_behavior_curriculum.task import SEMVER_REGEX, Task, TaskParameters

TTask = TypeVar("TTask", bound=Task)
_P = ParamSpec("_P")
_R = TypeVar("_R")

TTaskParameters = TypeVar("TTaskParameters", bound=TaskParameters)


class Metrics(AindBehaviorModelExtra):
    """
    Abstract Metrics class.
    Subclass with Metric values.
    """


TMetrics = TypeVar("TMetrics", bound=Metrics)


class _Rule(Generic[_P, _R]):
    """
    Custom Pydantic Type that defines de/serialization for Callables.
    This type is not intended to be used directly, but rather subclassed
    with explicit typing in the generics.

    The _Rule type wraps a Callable with a specific signature.
    The Callable type signature is evaluated at runtime to ensure that the
    type hinting matches the expected _Rule (or subtype) type signature.
    For the duck-type aficionados, this type can be skipped by calling
    Rule(..., skip_validation=True) or by not annotating the Callable.
    """

    def __init__(self, function: Callable[_P, _R], *, skip_validation: bool = False) -> None:
        """
        Initializes a new instance of the class.
        Args:
            function (Callable[_P, _R]): The function to be used. If an instance of _Rule is passed,
                                         the callable attribute of the _Rule instance will be used.
            skip_validation (bool, optional): If set to True, skips the validation of the callable's typing.
                                              Defaults to False.
        Returns:
            None
        """

        # Just in case people pass the _Rule instance directly
        # Instead of the callable
        if isinstance(function, _Rule):
            function = function.callable

        if not skip_validation:
            self._validate_callable_typing(function)
        self._callable = function

    def invoke(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        """Wraps the inner callable."""
        return self._callable(*args, **kwargs)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        """Wraps the inner callable."""
        # This is useful for structural typing in situations where we want to pass
        # _Rule as a callable flexibly
        return self.invoke(*args, **kwargs)

    def __eq__(self, other: Any) -> bool:
        """
        Custom equality method.
        Two instances of the same subclass type are considered equal.
        """
        if not isinstance(other, _Rule):
            return False
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """
        Returns the hash value for the object.
        """
        return hash(hash(self.name) + hash(self.callable))

    @property
    def name(self) -> str:
        """
        Name of the Rule.
        """
        return self.serialize_rule(self)

    @property
    def callable(self) -> Callable[_P, _R]:
        """Returns the wrapped callable."""
        return self._callable

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        """
        Custom validation, Ref:
        https://docs.pydantic.dev/latest/concepts/types/#handling-third-party-types
        """

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(cls._deserialize_rule),
            ]
        )

        from_callable_schema = core_schema.chain_schema(
            [
                core_schema.callable_schema(),
                core_schema.no_info_plain_validator_function(cls._deserialize_rule),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(cls),
                    from_str_schema,
                    from_callable_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(function=cls.serialize_rule),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _core_schema: core_schema.CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        """
        Custom validation, Ref:
        https://docs.pydantic.dev/latest/concepts/types/#as-a-method-on-a-custom-type
        """
        return handler(core_schema.str_schema())

    @classmethod
    def normalize_rule_or_callable(cls, rule: Callable | "_Rule") -> Self:
        """Ensures the outgoing type is normalized from a Callable or _Rule."""
        if isinstance(rule, cls):
            return rule
        if isinstance(rule, _Rule):
            return cls(rule.callable)
        if callable(rule):
            return cls(rule)
        else:
            raise TypeError("rule must be a Callable or _Rule type.")

    @classmethod
    def _deserialize_rule(cls, value: str | Callable[_P, _R]) -> "_Rule[_P, _R]":
        """
        Custom Deserialization.
        Imports function according to package and function name.
        """
        callable_handle: Callable[_P, _R]
        if callable(value):
            return cls(value)

        if isinstance(value, str):
            try:
                module_name, _, attr_path = value.partition(".")
                while "." in attr_path:
                    module_name += "." + attr_path.split(".", 1)[0]
                    attr_path = attr_path.split(".", 1)[1]

                module = importlib.import_module(module_name)
                obj = module
                for attr in attr_path.split("."):
                    obj = getattr(obj, attr)
            except (ModuleNotFoundError, AttributeError) as e:
                assert isinstance(value, str)
                callable_handle = _NonDeserializableCallable[_P, _R](value, e)
            else:
                callable_handle = cast(Callable[_P, _R], obj)

        return cls(callable_handle)

    @classmethod
    def serialize_rule(cls, value: Union[str, "_Rule[_P, _R]"]) -> str:
        """
        Custom Serialization.
        Simply exports reference to function as package + function name.
        """

        if isinstance(value, str):
            ret = cls._deserialize_rule(value)
        else:
            ret = value
        assert isinstance(ret, _Rule)

        if is_non_deserializable_callable(ret.callable):
            # Python 3.13 has a better way to infer
            # types with arbitrary clause code, but for now...
            assert isinstance(ret.callable, _NonDeserializableCallable)
            return ret.callable.mock_serialize()
        else:
            module = ret.callable.__module__
            qualname = ret.callable.__qualname__
            return f"{module}.{qualname}"

    @classmethod
    def _validate_callable_typing(cls, r: Callable[_P, _R]) -> None:
        """
        Validates that the provided callable `r` matches the expected signature
        defined by the generic types of the class.
        Args:
            r (Callable[_P, _R]): The callable to validate.
        Raises:
            ValueError: If `r` is not callable.
            TypeError: If the class does not define its generic types explicitly,
                    or if the callable `r` does not match the expected signature.
        """

        if isinstance(r, cls):
            return

        if is_non_deserializable_callable(r):
            return

        if not callable(r):
            raise ValueError("Rule must be callable.")

        # For some reason, generics do not materialize by default.
        # We fetch them manually....
        if isinstance(x := cls._solve_generic_typing(cls), TypeError):
            raise x
        else:
            expected_callable, expected_return = x

        # Compare signatures
        sig = inspect.signature(r)

        try:
            # Compare inputs
            if (_eval := cls._validate_signature_input(expected_callable, sig)) is not None:
                raise _eval

            if (_eval := cls._validate_signature_output(expected_return, sig)) is not None:
                raise _eval

        except TypeError as e:
            e.add_note(f"Expected callable type signature: {expected_callable} -> {expected_return}. Got {sig}.")
            raise e

    @staticmethod
    def _validate_signature_input(expected_callable: Any, sig: inspect.Signature) -> Optional[TypeError]:
        """Validates the input signature of the incoming callable against
        the expected input type."""

        if isinstance(expected_callable, EllipsisType):
            return None

        if not (len(sig.parameters) == 0 and expected_callable is type(None)):
            if not (len(sig.parameters) == len(expected_callable)):
                return TypeError(
                    f"Callable must have {len(expected_callable)} parameters, but {len(sig.parameters)} were found."
                )

            for param, expected_type in zip(list(sig.parameters.values()), expected_callable):
                if param.annotation is not param.empty:
                    if isinstance(expected_type, TypeVar):
                        expected_type = expected_type.__bound__
                        if expected_type is None:
                            return None  # We assume the Type is unbounded
                    if not issubclass(param.annotation, expected_type):
                        return TypeError(
                            f"Parameter '{param.name}' must be of type {expected_type}, but {param.annotation} was found."
                        )
        return None

    @staticmethod
    def _validate_signature_output(expected_return: Any, sig: inspect.Signature) -> Optional[TypeError]:
        """Validates the output signature of the incoming callable against
        the expected output type."""

        if isinstance(expected_return, EllipsisType):
            return None

        if sig.return_annotation is None and expected_return is type(None):
            return None

        if isinstance(expected_return, TypeVar):
            expected_return = expected_return.__bound__
            if expected_return is None:
                return None  # We assume the Type is unbounded

        if sig.return_annotation is not inspect.Signature.empty:
            if not issubclass(sig.return_annotation, expected_return):
                return TypeError(
                    f"Callable return type must be {expected_return}, but {sig.return_annotation} was found."
                )

        return None

    @staticmethod
    def _solve_generic_typing(cls_: Any) -> Tuple[Any, Any] | TypeError:
        """Returns an inner generic type's type hint information"""
        origin_bases = getattr(cls_, "__orig_bases__", [])
        if not origin_bases:
            return TypeError(f"Class {cls_.__name__} must define its generic types.")

        for base in origin_bases:
            # Hopefully nobody uses multiple inheritance with Rule
            # But just in case we grab the first Rule parent
            origin = get_origin(base)
            args = get_args(base)

            if origin is _Rule and len(args) == 2:
                # we assume that the constructor only takes a single
                # callable as argument. Can always be extended
                # partials with partial
                return (args[0], args[1])
        return TypeError(f"Class {cls_.__name__} must define its generic types explicitly.")


def is_non_deserializable_callable(value: Any) -> bool:
    """
    Check if the given value is an instance of _NonDeserializableCallable.
    Args:
        value (Any): The value to check.
    Returns:
        bool: True if the value is an instance of _NonDeserializableCallable, False otherwise.
    """

    return isinstance(value, _NonDeserializableCallable)


def try_materialize_non_deserializable_callable_error(
    value: "_NonDeserializableCallable",
) -> Optional[Exception]:
    """
    Attempts to materialize the error from a non-deserializable callable.

    Args:
        value (_NonDeserializableCallable): The value to check and potentially
                                            extract the error from.

    Returns:
        Optional[Exception]: The error associated with the non-deserializable
                             callable if it exists, otherwise None.
    """
    if not is_non_deserializable_callable(value):
        return None
    return value.error


class _NonDeserializableCallable(Generic[_P, _R]):
    """
    A class representing a reference to callable that could not be deserialized.
    """

    def __init__(self, callable_repr: str, error: Exception) -> None:
        """
        Initializes the instance with a callable representation and an error.

        Args:
            callable_repr (str): A string representation of the callable.
            error (Exception): The exception that was raised.

        Returns:
            None
        """
        self._callable_repr = callable_repr
        self._error = error

    @property
    def error(self) -> Exception:
        """
        Property that returns the error associated with the curriculum.

        Returns:
            Exception: The error associated with the curriculum.
        """
        return self._error

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        """Shim method to raise an error when the callable is called."""
        raise RuntimeError(f"Cannot call the non-deserializable callable reference '{self._callable_repr}'.")

    def mock_serialize(self) -> str:
        """Shim method to return the callable representation."""
        return self._callable_repr

    def __hash__(self):
        """Shim method to return the hash of the callable."""
        return hash(self._callable_repr)


class Policy(_Rule[[TMetrics, TTask], TTask], Generic[TMetrics, TTask]):
    """
    User-defined function that defines
    how the current Task is updated according to metrics.
    It subclasses _Rule.
    """

    pass


class PolicyTransition(_Rule[[TMetrics], bool], Generic[TMetrics]):
    """
    User-defined function that defines
    how current Policies change during a Stage.
    It subclasses _Rule.
    """

    pass


NodeTypes = TypeVar("NodeTypes")
EdgeType = TypeVar("EdgeType")


class _BehaviorGraph(AindBehaviorModel, Generic[NodeTypes, EdgeType]):
    """
    Core directed graph data structure used in Stage and Curriculum.
    """

    nodes: Dict[int, NodeTypes] = Field(default={}, validate_default=True)
    graph: Dict[int, List[Tuple[EdgeType, int]]] = Field(default={}, validate_default=True)

    def _get_node_id(self, node: NodeTypes) -> int:
        """
        Dictionaries are ordered for Python 3.7+ so this is safe.
        This library requires Python 3.8+.
        """
        dict_keys = list(self.nodes.keys())
        dict_values = list(self.nodes.values())

        i = dict_values.index(node)
        return dict_keys[i]

    def _create_node_id(self) -> int:
        """
        Helper method for add_node and add_transition.
        More readable than using hash(Policy).
        """
        new_id = 0
        if len(self.nodes) > 0:
            new_id = max(len(self.nodes), max(self.nodes.keys()) + 1)
        return new_id

    def add_node(self, node: NodeTypes) -> None:
        """
        Adds a floating node to the behavior graph.
        """
        p_id = self._create_node_id()
        self.nodes[p_id] = node
        self.graph[p_id] = []

    def remove_node(self, node: NodeTypes) -> None:
        """
        Removes node and all associated incoming/outgoing
        transition rules from the stage graph.
        NOTE: Removed nodes and transitions have the side effect
        of changing transition priority.
        """
        if node not in self.nodes.values():
            raise ValueError(f"Node {node} is not in the graph to be removed.")

        # Resolve node id
        p_id = self._get_node_id(node)

        # Remove node from node list
        del self.nodes[p_id]

        # Remove node from graph keys
        del self.graph[p_id]

        # Remove node from graph value lists
        for start_id in self.graph:
            for rule, dest_id in self.graph[start_id]:
                if dest_id == p_id:
                    self.graph[start_id].remove((rule, dest_id))

    def add_transition(
        self,
        start_node: NodeTypes,
        dest_node: NodeTypes,
        rule: EdgeType,
    ) -> None:
        """
        Add transition between two nodes:
        start_node -> dest_node.

        If start_node has been added to graph before, this method starts a transition
            from the existing start_node.
        If dest_node has been added to graph before, this method creates a transition
            into the existing dest_node.

        NOTE: The order in which this method
        is called sets the order of transition priority.
        """

        # Resolve id of start_node
        if start_node not in self.nodes.values():
            new_id = self._create_node_id()
            self.nodes[new_id] = start_node
            self.graph[new_id] = []
        start_id = self._get_node_id(start_node)

        # Resolve id of dest_node
        if dest_node not in self.nodes.values():
            new_id = self._create_node_id()
            self.nodes[new_id] = dest_node
            self.graph[new_id] = []
        dest_id = self._get_node_id(dest_node)

        # Add the new transition to the graph
        self.graph[start_id].append((rule, dest_id))

    def remove_node_transition(
        self,
        start_node: NodeTypes,
        dest_node: NodeTypes,
        rule: EdgeType,
        remove_start_node: bool = False,
        remove_dest_node: bool = False,
    ) -> None:
        """
        Removes transition with options to remove start/end nodes
        associated with the transition.
        NOTE: Removed nodes and transitions has the side effect
        of changing transition priority.
        """

        if start_node not in self.nodes.values():
            raise ValueError(f"Node {start_node} is not in the behavior graph to be removed.")

        if dest_node not in self.nodes.values():
            raise ValueError(f"Node {dest_node} is not in the behavior graph to be removed.")

        start_id = self._get_node_id(start_node)
        dest_id = self._get_node_id(dest_node)
        if (rule, dest_id) not in self.graph[start_id]:
            raise ValueError(f"Node {start_node} does not transition into Node {dest_node} with Rule {rule}.")

        # Optionally remove nodes
        if remove_start_node:
            self.remove_node(start_node)
        if remove_dest_node:
            self.remove_node(dest_node)

        # Remove transition
        self.graph[start_id].remove((rule, dest_id))

    def see_nodes(self) -> List[NodeTypes]:
        """
        See nodes of behavior graph.
        """
        return list(self.nodes.values())

    def see_node_transitions(self, node: NodeTypes) -> List[Tuple[EdgeType, NodeTypes]]:
        """
        See transitions of node in behavior graph.
        """

        if node not in self.nodes.values():
            raise ValueError(f"Node {node} is not in the behavior graph.")

        node_id = self._get_node_id(node)
        node_list = self.graph[node_id]
        return [(rule, self.nodes[p_id]) for (rule, p_id) in node_list]

    def set_transition_priority(
        self,
        node: NodeTypes,
        node_transitions: List[Tuple[EdgeType, NodeTypes]],
    ) -> None:
        """
        Change order of node transitions listed under a node.
        Highest priority is ordered in node_transition from left -> right.
        """

        input_transitions = []
        for rule, n in node_transitions:
            if n not in self.nodes.values():
                raise ValueError(f"Node {n} is not a node inside the behavior graph.")
            input_transitions.append((rule, self._get_node_id(n)))

        n_id = self._get_node_id(node)
        self.graph[n_id] = input_transitions

    def __eq__(self, other: Any) -> bool:
        """
        Compare this object with another for equality.

        Args:
            other (Any): The object to compare with.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        if not isinstance(other, _BehaviorGraph):
            return False
        return self.model_dump() == other.model_dump()


class PolicyGraph(_BehaviorGraph[Policy[TMetrics, TTask], PolicyTransition[TMetrics]], Generic[TMetrics, TTask]):
    """
    Graph for Stage.
    """

    pass


class MetricsProvider(_Rule[..., TMetrics], Generic[TMetrics]):
    """A type for a callable that is able to produce Metrics"""

    pass


class Stage(AindBehaviorModel, Generic[TMetrics, TTask]):
    """
    Instance of a Task.
    Task Parameters may change according to rules defined in BehaviorGraph.
    Stage manages a BehaviorGraph instance with a read/write API.
    """

    name: str = Field(..., description="Stage name.")
    task: TTask = Field(..., description="Task in which this stage is based off of.")
    graph: PolicyGraph[TMetrics, TTask] = Field(
        default_factory=PolicyGraph[TMetrics, TTask],
        validate_default=True,
        description="Policy Graph.",
    )
    start_policies: List[Policy[TMetrics, TTask]] = Field(
        default_factory=list, description="List of starting policies."
    )
    metrics_provider: Optional[MetricsProvider[TMetrics]] = Field(
        default=None,
        description="A MetricsProvider instance that keeps a reference to a handle to create a metrics object for this stage.",
    )

    def __eq__(self, other: Any) -> bool:
        """
        Custom equality method.
        Two Stage instances are only distinguished by name.
        """
        # TODO. We should consider cleaning this up in the future.
        # Since Stage is mutable at the level of the TaskParameters
        # we cannot simply compare the model_dump() of the two instances.
        # However, we need the __eq__ to look for nodes in the graph.
        # The solution is probably to make our own "look_up" method
        # This should generally be safe since the Stage name is unique,
        # but it is a brittle solution.
        if not isinstance(other, Stage):
            return False
        return self.name == other.name

    def model_post_init(self, __context):
        """Runs after model_construct to ensure that the
        initial policies update the PolicyGraph"""
        super().model_post_init(__context)
        self.set_start_policies(self.start_policies, append_non_existing=True)

    def set_start_policies(
        self,
        start_policies: Policy[TMetrics, TTask] | Iterable[Policy[TMetrics, TTask]],
        append_non_existing: bool = True,
    ) -> None:
        """
        Sets stage's start policies to start policies provided.
        Input overwrites existing start policies.
        """
        if not isinstance(start_policies, Iterable):
            start_policies = [start_policies]

        for policy in start_policies:
            policy = Policy.normalize_rule_or_callable(policy)
            if policy not in self.graph.see_nodes():
                if append_non_existing:
                    self.add_policy(policy)
                else:
                    raise ValueError(f"Policy {policy} is not in the policy graph.")

        self.start_policies = list(start_policies)

    def add_policy(self, policy: Policy[TMetrics, TTask]) -> None:
        """
        Adds a floating policy to the Stage adjacency graph.
        """
        policy = Policy.normalize_rule_or_callable(policy)
        if policy in self.graph.see_nodes():
            raise ValueError(f"Policy {policy.name} is a duplicate Policy in Stage {self.name}.")

        self.graph.add_node(policy)

    def remove_policy(self, policy: Policy[TMetrics, TTask]) -> None:
        """
        Removes policy and all associated incoming/outgoing
        transition rules from the stage graph.
        NOTE: Removed nodes and transitions have the side effect
        of changing transition priority.
        """
        policy = Policy.normalize_rule_or_callable(policy)
        self.graph.remove_node(policy)

        # Also remove reference to policy in start_policies if applicable.
        if policy in self.start_policies:
            self.start_policies.remove(policy)

            if len(self.start_policies) == 0:
                warnings.warn(f"Stage {self.name}.start_policies is empty.")

    def add_policy_transition(
        self,
        start_policy: Policy[TMetrics, TTask],
        dest_policy: Policy[TMetrics, TTask],
        rule: PolicyTransition[TMetrics],
    ) -> None:
        """
        Add policy transition between two policies:
        Policy_A -> Policy_B.

        If Policy_A has been added to stage before, this method starts a transition
            from the existing Policy_A.
        If Policy_B has been added to stage before, this method creates a transition
            into the existing Policy_B.

        NOTE: The order in which this method
        is called sets the order of transition priority.
        """

        self.graph.add_transition(
            Policy.normalize_rule_or_callable(start_policy),
            Policy.normalize_rule_or_callable(dest_policy),
            PolicyTransition.normalize_rule_or_callable(rule),
        )

    def remove_policy_transition(
        self,
        start_policy: Policy[TMetrics, TTask],
        dest_policy: Policy[TMetrics, TTask],
        rule: PolicyTransition[TMetrics],
        remove_start_policy: bool = False,
        remove_dest_policy: bool = False,
    ) -> None:
        """
        Removes transition with options to remove start/end policies
        associated with the transition.
        NOTE: Removed nodes and transitions has the side effect
        of changing transition priority.
        """
        self.graph.remove_node_transition(
            Policy.normalize_rule_or_callable(start_policy),
            Policy.normalize_rule_or_callable(dest_policy),
            PolicyTransition.normalize_rule_or_callable(rule),
            remove_start_policy,
            remove_dest_policy,
        )

    def see_policies(self) -> List[Policy[TMetrics, TTask]]:
        """
        See policies of policy graph.
        """
        return self.graph.see_nodes()

    def see_policy_transitions(
        self, policy: Policy[TMetrics, TTask]
    ) -> List[Tuple[PolicyTransition[TMetrics], Policy[TMetrics, TTask]]]:
        """
        See transitions of stage in policy graph.
        """

        return self.graph.see_node_transitions(Policy.normalize_rule_or_callable(policy))

    def set_policy_transition_priority(
        self,
        policy: Policy[TMetrics, TTask],
        policy_transitions: List[Tuple[PolicyTransition[TMetrics], Policy[TMetrics, TTask]]],
    ) -> None:
        """
        Change the order of policy transitions listed under a policy.
        To use, call see_policy_transitions() and order the transitions
        in the desired priority from left -> right.
        """
        policy = Policy.normalize_rule_or_callable(policy)
        policy_transitions = [
            (
                PolicyTransition.normalize_rule_or_callable(t),
                Policy.normalize_rule_or_callable(p),
            )
            for (t, p) in policy_transitions
        ]
        current_list = [(t, p) for (t, p) in self.see_policy_transitions(policy)]

        if len(policy_transitions) != len(current_list):
            raise ValueError(
                f"Number of input node transitions {policy_transitions} does not \
                match the number of elements under this node: {self.see_policy_transitions(policy)}.\
                Please call 'see_policy_transitions()' for a precise list of elements."
            )

        self.graph.set_transition_priority(policy, policy_transitions)

    def get_task(self) -> TTask:
        """
        Get the current task as a deep-copy for safe manipulation.
        """
        return self.task.model_copy(deep=True)

    def set_task(self, task: TTask) -> None:
        """
        Set the current task using a copy of the input.
        """
        self.task = task.model_copy(deep=True)

    @deprecated("This method is deprecated in favor of setting the task directly using set_task(...).")
    def set_task_parameters(self, task_parameters: TTaskParameters) -> None:
        """
        Set the task parameters for the current task.
        """
        self.task = self.get_task().model_copy(update={"task_parameters": task_parameters})

    @deprecated("This method is deprecated in favor of getting the task directly using get_task(...).")
    def get_task_parameters(self) -> TTaskParameters:
        """
        Get the task parameters for the current task.
        """
        return self.get_task().task_parameters

    def validate_stage(self) -> Self:
        """
        Validates that the stage can be (de)serialized.
        """

        # Check round trip serialization
        try:
            instance_json = self.model_dump_json()
            self.model_validate_json(instance_json)
        except ValidationError as e:
            e.add_note(
                f"Pydantic cannot serialize Stage {self.name}, please use "
                "mypy to verify your types "
                "(check signatures of policy functions, etc.)."
            )
            raise e

        return self


class StageTransition(_Rule[[TMetrics], bool], Generic[TMetrics]):
    """
    User-defined function that defines
    criteria for transitioning stages based on metrics.
    Subclasses _Rule.
    """

    pass


class StageGraph(_BehaviorGraph[Stage[TMetrics, TTask], StageTransition[TMetrics]], Generic[TMetrics, TTask]):
    """
    Graph for Curriculum.
    """

    pass


class Curriculum(AindBehaviorModel, Generic[TTask]):
    """
    Curriculum manages a StageGraph instance with a read/write API.
    To use, subclass this and add subclass metrics.
    """

    pkg_location: str = Field(
        # https://docs.python.org/3/faq/programming.html#why-do-lambdas-defined-in-a-loop-with-different-values-all-return-the-same-result
        default_factory=lambda: Curriculum.default_pkg_location_factory(),  # pylint: disable=unnecessary-lambda
        frozen=True,
        description="Location of the python package \
                                that instantiated the Curriculum.",
    )
    name: str = Field(
        default="Please subclass, rename, and define \
                      a StageGraph with your own Stage objs \
                      Ex: StageGraph[Union[StageA, StageB, Graduated]]",
        description="Name of the Curriculum.",
        frozen=True,
    )
    version: str = Field(
        default="0.0.0",
        pattern=SEMVER_REGEX,
        frozen=True,
        validate_default=True,
        description="Curriculum version.",
    )
    graph: StageGraph[Metrics, TTask] = Field(default_factory=StageGraph[Metrics, TTask], validate_default=True)

    @property
    def _known_tasks(self) -> List[Type[Task]]:
        """Get all known tasks in the curriculum."""

        # We introspect into the StageGraph[T] type to get the known tasks.
        _generic = self.model_fields["graph"].annotation
        _inner_args = _generic.__dict__["__pydantic_generic_metadata__"]["args"]

        _inner_union: Type
        if len(_inner_args) == 0:
            _inner_union = Task
        else:
            # TODO This may be an issue if people define a generic on TTask but not Metrics
            _inner_args = _inner_args[1]
            _inner_union = get_args(_inner_args.__value__)[0]

        if isinstance(_inner_union, type):
            _known_tasks = [_inner_union]
        else:
            _known_tasks = [x for x in get_args(_inner_union)]

        # Since we are here, we also check if the known tasks match the nodes in the graph
        # The tasks known to the graph type should be a super set of the known tasks in the nodes
        _known_nodes = [type(stage.task) for stage in self.graph.see_nodes()]
        if not set(_known_tasks).issuperset(set(_known_nodes)):
            raise ValueError(
                "Known tasks in the Curriculum do not match the tasks in the StageGraph. This is likely a problem with StageGraph type definition."
            )
        return _known_tasks

    def task_discriminator_type(self) -> Type[Task]:
        """Create a Discriminated Union  type for the known tasks."""
        return make_task_discriminator(self._known_tasks)

    def _is_task_type_known(self, task_type: Task | Type[Task]) -> bool:
        """Check if a task type is known in the curriculum."""
        if isinstance(task_type, Task):
            task_type = type(task_type)
        if isinstance(task_type, type):
            return task_type in self._known_tasks
        raise ValueError("task_type must be a Task instance or a Task type.")

    @classmethod
    def default_pkg_location_factory(cls) -> str:
        """
        Location of the python package that instantiated the Curriculum.
        """
        return cls.__module__ + "." + type(cls).__name__

    def add_stage(self, stage: Stage) -> None:
        """
        Adds a floating stage to the Curriculum adjacency graph.
        """
        task = stage.get_task()
        if not self._is_task_type_known(task):
            raise ValueError(f"Task {task} is not a known task type in the Curriculum.")

        if stage in self.graph.see_nodes():
            raise ValueError(f"Stage {stage.name} is a duplicate stage in Curriculum.")

        self.graph.add_node(stage)

    def remove_stage(self, stage: Stage) -> None:
        """
        Removes stage and all associated incoming/outgoing
        transition rules from the curriculum graph.
        NOTE: Removed nodes and transitions have the side effect
        of changing transition priority.
        """
        self.graph.remove_node(stage)

    def add_stage_transition(
        self,
        start_stage: Stage,
        dest_stage: Stage,
        rule: StageTransition,
    ) -> None:
        """
        Add stage transition between two stages:
        Stage_A -> Stage_B.

        If Stage_A has been added to stage before, this method starts a transition
            from the existing Stage_A.
        If Stage_B has been added to stage before, this method creates a transition
            into the existing Stage_B.

        NOTE: The order in which this method
        is called sets the order of transition priority.
        """

        self.graph.add_transition(
            start_stage,
            dest_stage,
            StageTransition.normalize_rule_or_callable(rule),
        )

    def remove_stage_transition(
        self,
        start_stage: Stage,
        dest_stage: Stage,
        rule: StageTransition,
        remove_start_stage: bool = False,
        remove_dest_stage: bool = False,
    ) -> None:
        """
        Removes transition with options to remove start/end stages
        associated with the transition.
        NOTE: Removed nodes and transitions has the side effect
        of changing transition priority.
        """
        self.graph.remove_node_transition(
            start_stage,
            dest_stage,
            StageTransition.normalize_rule_or_callable(rule),
            remove_start_stage,
            remove_dest_stage,
        )

    def see_stages(self) -> List[Stage]:
        """
        See stages of curriculum graph.
        """
        return self.graph.see_nodes()

    def see_stage_transitions(self, stage: Stage) -> List[Tuple[StageTransition, Stage]]:
        """
        See transitions of stage in curriculum graph.
        """
        return self.graph.see_node_transitions(stage)

    def set_stage_transition_priority(
        self,
        stage: Stage,
        stage_transitions: List[Tuple[StageTransition, Stage]],
    ) -> None:
        """
        Change the order of stage transitions listed under a stage.
        To use, call see_stage_transitions() and order the transitions
        in the desired priority from left -> right.
        """

        stage_transitions = [(StageTransition.normalize_rule_or_callable(t), s) for (t, s) in stage_transitions]

        if len(stage_transitions) != len(self.see_stage_transitions(stage)):
            raise ValueError(
                f"Elements of input node transitions {stage_transitions} does not \
                    match the elements under this node: {self.see_stage_transitions(stage)}. \
                        Please call 'see_stage_transitions()' for a precise list of elements."
            )

        self.graph.set_transition_priority(stage, stage_transitions)

    def validate_curriculum(self) -> Self:
        """
        Validate curriculum for export/serialization.
        """

        if not all([self._is_task_type_known(stage.task) for stage in self.see_stages()]):
            raise ValueError("Not all tasks in the curriculum are known. Please add stages with known tasks.")

        if len(self.see_stages()) == 0:
            raise ValueError("Curriculum is empty! Please add stages.")

        for s in self.see_stages():
            s.validate_stage()

        # Check round trip serialization
        try:
            instance_json = self.model_dump_json()
            self.model_validate_json(instance_json)
        except ValidationError as e:
            e.add_note(
                (
                    "Pydantic cannot serialize Curriculum, please use "
                    "mypy to verify your types (check stage transition signature, etc.)."
                )
            )
            raise e

        return self


def create_curriculum(
    name: str,
    version: str,
    tasks: Iterable[Type[TTask]],
    pkg_location: Optional[str] = None,
) -> Type[Curriculum[TTask]]:
    """
    Creates a new curriculum model with the specified name, version, and tasks.
    Args:
        name (str): The name of the curriculum.
        version (str): The version of the curriculum, following semantic versioning.
        tasks (Iterable[Type[Task]]): An iterable of Task types to be included in the curriculum.
        pkg_location (Optional[str]): Optional package location string.
    Returns:
        Type[Curriculum]: A new curriculum model type.
    Raises:
        ValueError: If no tasks are provided.
    """

    if not any(tasks):
        raise ValueError("At least one task must be provided.")

    _tasks_tagged = cast(TTask, make_task_discriminator(tasks))

    if pkg_location is not None:
        return create_model(
            name,
            __base__=Curriculum[_tasks_tagged],
            name=Annotated[
                Literal[name],
                Field(default=name, frozen=True, validate_default=True),
            ],
            version=Annotated[
                Literal[version] if version else Optional[str],
                Field(
                    default=version,
                    frozen=True,
                    pattern=SEMVER_REGEX,
                    validate_default=True,
                ),
            ],
            pkg_location=Annotated[
                str,
                Field(default=pkg_location, frozen=False, validate_default=True),
            ],
            graph=Annotated[
                StageGraph[Metrics, _tasks_tagged],
                Field(default_factory=StageGraph[Metrics, _tasks_tagged], validate_default=True),
            ],
        )
    else:
        return create_model(
            name,
            __base__=Curriculum[_tasks_tagged],
            name=Annotated[
                Literal[name],
                Field(default=name, frozen=True, validate_default=True),
            ],
            version=Annotated[
                Literal[version] if version else Optional[str],
                Field(
                    default=version,
                    frozen=True,
                    pattern=SEMVER_REGEX,
                    validate_default=True,
                ),
            ],
            graph=Annotated[
                StageGraph[Metrics, _tasks_tagged],
                Field(default_factory=StageGraph[Metrics, _tasks_tagged], validate_default=True),
            ],
        )


def make_task_discriminator(tasks: Iterable[Type[TTask]]) -> Type[TTask]:
    """
    Creates a discriminated union type for the given tasks.
    This function takes a variable number of Task types and generates a
    discriminated union type using the 'name' field of each task to create
    a discriminated union.
    Args:
        tasks (Iterable[Type[Task]]): A variable number of Task types.
    Returns:
        Type: A TypeAliasType with the discriminated union type of the provided tasks.
    """

    return cast(
        Type[TTask],
        TypeAliasType(
            "known_task_types",
            Annotated[
                Union[tuple(set(tasks))],
                Field(discriminator="name"),
            ],
        ),
    )



"""
Init package

Modules in this file are public facing.
"""

__version__ = "0.0.35"

from .curriculum import (
    Curriculum,
    Metrics,
    MetricsProvider,
    Policy,
    PolicyTransition,
    Stage,
    StageGraph,
    StageTransition,
    create_curriculum,
    make_task_discriminator,
)
from .curriculum_utils import GRADUATED, Graduated, export_diagram, export_json
from .task import Task, TaskParameters, create_task
from .trainer import Trainer, TrainerServer, TrainerState

__all__ = [
    "Curriculum",
    "Metrics",
    "MetricsProvider",
    "Policy",
    "PolicyTransition",
    "Stage",
    "StageGraph",
    "StageTransition",
    "create_curriculum",
    "make_task_discriminator",
    "GRADUATED",
    "Graduated",
    "Task",
    "TaskParameters",
    "create_task",
    "Trainer",
    "TrainerServer",
    "TrainerState",
    "export_diagram",
    "export_json",
]



"""
Example of Trainer creation
"""

from collections import defaultdict
from typing import Dict

import example_project as ex
from aind_behavior_curriculum import (
    Curriculum,
    Metrics,
    TrainerServer,
    TrainerState,
)

# Proxy Database
# NOTE: Trainer's concrete implementation
# assumes a higher-level process defines mouse ID's ahead of time
MICE_CURRICULUMS: dict[int, Curriculum] = {}
MICE_SUBJECT_HISTORY: dict[int, list[TrainerState]] = defaultdict(list)
MICE_METRICS: dict[int, Metrics] = {
    0: ex.ExampleMetrics(),
    1: ex.ExampleMetrics(),
    2: ex.ExampleMetrics(),
}


class ExampleTrainer(TrainerServer):
    def __init__(self) -> None:
        """
        Custom init w/ super.__init__()
        Add database connections, etc. here
        """
        super().__init__()

        self.subject_history: Dict[int, TrainerState] = defaultdict(list)

    def load_data(self, subject_id: int) -> tuple[Curriculum, TrainerState, Metrics]:
        """
        Read from proxy database.
        """
        return (
            MICE_CURRICULUMS[subject_id],
            MICE_SUBJECT_HISTORY[subject_id][-1],
            MICE_METRICS[subject_id],
        )

    def write_data(
        self,
        subject_id: int,
        curriculum: Curriculum,
        trainer_state: TrainerState,
    ) -> None:
        """
        Add to proxy database.
        """
        MICE_CURRICULUMS[subject_id] = curriculum
        MICE_SUBJECT_HISTORY[subject_id].append(trainer_state)

        self.subject_history[subject_id].append(trainer_state)



"""
Base behavior pydantic object
"""

from pydantic import BaseModel, ConfigDict


class AindBehaviorModel(BaseModel):
    """
    Defines Pydantic configurations applied to all behavior models.
    BaseModel: Validate arguments on initialization.
    Configurations:

    - **extra**: 'forbid'
        Do not allow a model to be initialized with undocumented parameters.
    - **validate_assignment**: True
        Revalidate fields against schema on any change to model instance.
    - **validate_default**: True
        Validate default fields on subclasses.
    - **strict**: True
        Enforce strict typing.
    """

    model_config = ConfigDict(
        extra="forbid",  # Potentially change to 'ignore'
        validate_assignment=True,
        validate_default=True,
        strict=True,
        str_strip_whitespace=True,
    )


class AindBehaviorModelExtra(BaseModel):
    """
    Same as AindBehaviorModel w/ extra = "allow".
    Helpful for deserialization of nested subclasses.
    """

    model_config = ConfigDict(
        extra="allow",
        validate_assignment=True,
        validate_default=True,
        strict=True,
        str_strip_whitespace=True,
    )



"""
Task Test Suite
"""

import unittest

import example_project as ex

from aind_behavior_curriculum import Task


class TaskTests(unittest.TestCase):
    """Unit tests for valid and invalid Task usage"""

    # Valid Usage
    def test_valid_construction(self):
        ex_parameters = ex.ExampleTaskParameters(field_2=50, field_4=0.8)
        ex_task = ex.ExampleTask(task_parameters=ex_parameters)
        self.assertEqual(ex_task.task_parameters.field_2, 50)
        self.assertEqual(ex_task.task_parameters.field_4, 0.8)

    def test_valid_parameter_change(self):
        ex_parameters = ex.ExampleTaskParameters()
        ex_task = ex.ExampleTask(task_parameters=ex_parameters)
        ex_task.task_parameters.field_1 = 50
        self.assertEqual(ex_task.task_parameters.field_1, 50)

    # Invalid Usage
    def test_invalid_construction(self):
        def invalid_type():
            ex_parameters = ex.ExampleTaskParameters(field_1="20")
            ex_task = ex.ExampleTask(  # noqa: F841
                task_parameters=ex_parameters
            )

        def invalid_field():
            ex_parameters = ex.ExampleTaskParameters(field_4=5)
            ex_task = ex.ExampleTask(  # noqa: F841
                task_parameters=ex_parameters
            )

        self.assertRaises(Exception, invalid_type)
        self.assertRaises(Exception, invalid_field)

    def test_invalid_parameter_change(self):
        def invalid_type():
            ex_parameters = ex.ExampleTaskParameters()
            ex_task = ex.ExampleTask(task_parameters=ex_parameters)
            ex_task.update_parameters(field_1="20")

        def invalid_field():
            ex_parameters = ex.ExampleTaskParameters()
            ex_task = ex.ExampleTask(task_parameters=ex_parameters)
            ex_task.update_parameters(field_4=5)

        self.assertRaises(Exception, invalid_type)
        self.assertRaises(Exception, invalid_field)

    def test_round_trip(self):
        ex_parameters = ex.ExampleTaskParameters()
        ex_task = ex.ExampleTask(task_parameters=ex_parameters)

        # Serialize from Child
        instance_json = ex_task.model_dump_json()
        # Deserialize from Child
        recovered = ex.ExampleTask.model_validate_json(instance_json)
        self.assertEqual(ex_task, recovered)

        # Serialize from Child
        instance_json = ex_task.model_dump_json()
        # Deserialize from Parent
        instance_parent = Task.model_validate_json(instance_json)
        # Serialize from Parent
        parent_json = instance_parent.model_dump_json()
        # Deserialize from Child
        instance_prime = ex.ExampleTask.model_validate_json(parent_json)
        self.assertEqual(ex_task, instance_prime)


if __name__ == "__main__":
    unittest.main()



"""
Trainer Test Suite
"""

import unittest

import example_project as ex
import example_project_2 as ex2

from aind_behavior_curriculum import GRADUATED, Stage, Trainer, TrainerState


class TrainerTests(unittest.TestCase):
    def test_create_enrollment(self):
        """Tests the syntactic sugar method for creating a initial
        state for the trainer"""
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())
        stageA = Stage(name="StageA", task=taskA)
        stageB = Stage(name="StageB", task=taskB)

        curr = ex.MyCurriculum(name="My Curriculum")
        curr.add_stage_transition(stageA, GRADUATED, ex.t2_10)
        curr.add_stage_transition(stageA, stageB, ex.t2_5)
        curr.add_stage_transition(stageB, GRADUATED, ex.t2_10)

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA)
        tr.register_subject(1, curr, stageA)
        tr.register_subject(2, curr, stageA)

        # Move mice through curriculum by updating metrics
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=0)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=6)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        enrollment = trainer.create_enrollment()
        expected = trainer.create_trainer_state(
            stage=stageA, is_on_curriculum=True, active_policies=stageA.start_policies
        )
        self.assertEqual(enrollment, expected)
        self.assertEqual(enrollment.model_dump(), expected.model_dump())

    def test_pure_stage_evaluation(self):
        """
        Tests if multiple trajectories through stages
        are recorded correctly in database.
        """

        # Create Stage-only curriculum
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())
        stageA = Stage(name="StageA", task=taskA)
        stageB = Stage(name="StageB", task=taskB)

        curr = ex.MyCurriculum(name="My Curriculum")
        curr.add_stage_transition(stageA, GRADUATED, ex.t2_10)
        curr.add_stage_transition(stageA, stageB, ex.t2_5)
        curr.add_stage_transition(stageB, GRADUATED, ex.t2_10)

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA)
        tr.register_subject(1, curr, stageA)
        tr.register_subject(2, curr, stageA)

        # Move mice through curriculum by updating metrics
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=0)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=6)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        # Validate mouse histories
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[]),
            trainer.create_trainer_state(stage=stageA, active_policies=[]),
            trainer.create_trainer_state(stage=stageB, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        M1 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[]),
            trainer.create_trainer_state(stage=stageB, active_policies=[]),
            trainer.create_trainer_state(stage=stageB, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        M2 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        self.assertEqual(tr.subject_history[0], M0)
        self.assertEqual(tr.subject_history[1], M1)
        self.assertEqual(tr.subject_history[2], M2)

    def test_pure_policy_evaluation(self):
        """
        Tests if multiple trajectories through policies
        are recorded correctly in database, including
        stage parameter changes by policy updates.
        """

        # Create single-stage curriculum
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        stageA = Stage(name="StageA", task=taskA)
        stageA.add_policy_transition(ex.INIT_STAGE, ex.stageA_policyB, ex.t1_10)
        stageA.add_policy_transition(ex.INIT_STAGE, ex.stageA_policyA, ex.t1_5)
        stageA.add_policy_transition(ex.stageA_policyA, ex.stageA_policyB, ex.t1_10)
        stageA.set_start_policies(ex.INIT_STAGE)

        stageAA = stageA.model_copy(deep=True)
        stageAA.set_task(ex.stageA_policyA_rule(ex.ExampleMetrics(), taskA.model_copy()))

        stageAAA = stageA.model_copy(deep=True)
        stageAAA.set_task(ex.stageA_policyB_rule(ex.ExampleMetrics(), taskA.model_copy()))

        curr = ex.MyCurriculum(name="My Curriculum")
        curr.add_stage(stageA)

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA, ex.INIT_STAGE)
        tr.register_subject(1, curr, stageA, ex.INIT_STAGE)
        tr.register_subject(2, curr, stageA, ex.INIT_STAGE)

        # Move mice through curriculum by updating metrics
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=0)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_1=12)
        tr.evaluate_subjects()

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=6)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_1=12)
        tr.evaluate_subjects()

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=12)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_1=12)
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        # Validate mouse histories
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAA, active_policies=[ex.stageA_policyA]),
            trainer.create_trainer_state(stage=stageAA, active_policies=[ex.stageA_policyB]),
        ]

        M1 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAA, active_policies=[ex.stageA_policyA]),
            trainer.create_trainer_state(stage=stageAA, active_policies=[ex.stageA_policyA]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
        ]

        M2 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
        ]

        self.assertEqual(tr.subject_history[0], M0)
        self.assertEqual(tr.subject_history[1], M1)
        self.assertEqual(tr.subject_history[2], M2)

    def test_stage_policy_evaluation(self):
        """
        Tests if multiple trajectories through stages and policies
        are recorded correctly in database.
        """
        curr = ex.construct_curriculum()

        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())

        # See stages is ordered arbitrarily
        # StageA and StageB picked out from observation
        stageA = curr.see_stages()[0]
        stageB = curr.see_stages()[2]
        stageA.set_start_policies([ex.INIT_STAGE])
        stageB.set_start_policies([ex.INIT_STAGE])

        stageAA = stageA.model_copy(deep=True)
        stageAA.set_task(ex.stageA_policyA_rule(ex.ExampleMetrics(), taskA.model_copy()))
        stageAAA = stageA.model_copy(deep=True)
        stageAAA.set_task(ex.stageA_policyB_rule(ex.ExampleMetrics(), taskA.model_copy()))
        stageBB = stageB.model_copy(deep=True)

        stageBB.set_task(ex.stageB_policyA_rule(ex.ExampleMetrics(), taskB.model_copy()))
        stageBBB = stageB.model_copy(deep=True)
        stageBBB.set_task(ex.stageB_policyB_rule(ex.ExampleMetrics(), taskB.model_copy()))

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA)
        tr.register_subject(1, curr, stageA)
        tr.register_subject(2, curr, stageA)

        # Move mice through curriculum by updating metrics
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=8)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_1=12)
        tr.evaluate_subjects()

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=8)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_2=12)
        tr.evaluate_subjects()

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_3=8)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_3=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_3=12)
        tr.evaluate_subjects()

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_2=12)
        ex.MICE_METRICS[2] = ex.ExampleMetrics(theta_3=12)
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        # Validate mouse histories
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAA, active_policies=[ex.stageA_policyA]),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageBB, active_policies=[ex.stageB_policyA]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        M1 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageBBB, active_policies=[ex.stageB_policyB]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        M2 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        self.assertEqual(tr.subject_history[0], M0)
        self.assertEqual(tr.subject_history[1], M1)
        self.assertEqual(tr.subject_history[2], M2)

    def test_status_override(self):
        """
        Test if status override
        is recorded correctly in database.
        """

        curr = ex.construct_curriculum()

        # See stages is ordered arbitrarily
        # StageA and StageB picked out from observation
        stageA = curr.see_stages()[0]
        stageB = curr.see_stages()[2]
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())
        stageAA = stageA.model_copy(deep=True)
        stageAA.set_task(ex.stageA_policyA_rule(ex.ExampleMetrics(), taskA.model_copy()))
        stageAAA = stageA.model_copy(deep=True)
        stageAAA.set_task(ex.stageA_policyB_rule(ex.ExampleMetrics(), taskA.model_copy()))
        stageBB = stageB.model_copy(deep=True)
        stageBB.set_task(ex.stageB_policyA_rule(ex.ExampleMetrics(), taskB.model_copy()))
        stageBBB = stageB.model_copy(deep=True)
        stageBBB.set_task(ex.stageB_policyB_rule(ex.ExampleMetrics(), taskB.model_copy()))

        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA, ex.INIT_STAGE)

        # Override API
        tr.override_subject_status(0, override_stage=stageBB, override_policies=ex.stageB_policyA)
        tr.override_subject_status(0, override_stage=stageAAA, override_policies=ex.stageA_policyB)
        tr.override_subject_status(0, override_stage=GRADUATED, override_policies=())
        trainer = Trainer(curr)
        # Validate mouse history
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageBB, active_policies=[ex.stageB_policyA]),
            trainer.create_trainer_state(stage=stageAAA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=GRADUATED, active_policies=[]),
        ]

        self.assertEqual(tr.subject_history[0], M0)

    def test_floating_stages(self):
        """
        Test override between in-curriculum floating stages.
        Stages have no transitions, thus SubjectHistory should
        remain static until there is an override.
        """

        # Create Stage-only curriculum
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())
        stageA = Stage(name="StageA", task=taskA)
        stageA.set_start_policies(ex.INIT_STAGE, append_non_existing=True)
        stageB = Stage(name="StageB", task=taskB)
        stageB.set_start_policies(ex.INIT_STAGE, append_non_existing=True)

        curr = ex.MyCurriculum(name="My Curriculum")
        curr.add_stage(stageA)
        curr.add_stage(stageB)

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA, ex.INIT_STAGE)
        tr.register_subject(1, curr, stageB, ex.INIT_STAGE)

        # Regular evaluation logs same entry
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=10, theta_2=10)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=10, theta_2=10)
        tr.evaluate_subjects()

        # Only way to move mouse is with override
        tr.override_subject_status(0, override_stage=stageB, override_policies=stageB.start_policies)
        tr.override_subject_status(1, override_stage=stageA, override_policies=stageA.start_policies)
        tr.override_subject_status(0, override_stage=stageA, override_policies=stageA.start_policies)
        tr.override_subject_status(1, override_stage=stageB, override_policies=stageB.start_policies)

        trainer = Trainer(curr)
        # Validate mouse history
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
        ]

        M1 = [
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
        ]

        self.assertEqual(tr.subject_history[0], M0)
        self.assertEqual(tr.subject_history[1], M1)

    def test_floating_policies(self):
        """
        Test policy override within a stage.
        Policies have no transitions, thus SubjectHistory should
        remain static until there is an override.
        """

        # Create single-stage curriculum
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        stageA = Stage(name="StageA", task=taskA)

        stageA.add_policy(ex.INIT_STAGE)
        stageA.add_policy(ex.stageA_policyA)
        stageA.add_policy(ex.stageA_policyB)
        stageA.set_start_policies(ex.INIT_STAGE)

        stageAA = stageA.model_copy(deep=True)
        stageAA.set_task(ex.stageA_policyA_rule(ex.ExampleMetrics(), taskA.model_copy()))
        stageAAA = stageA.model_copy(deep=True)
        stageAAA.set_task(ex.stageA_policyB_rule(ex.ExampleMetrics(), taskA.model_copy()))

        curr = ex.MyCurriculum(name="My Curriculum")
        curr.add_stage(stageA)

        # Associate mice with curriculum
        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA, ex.INIT_STAGE)
        tr.register_subject(1, curr, stageA, ex.INIT_STAGE)

        # Regular evaluation logs same entry
        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=10, theta_2=10)
        ex.MICE_METRICS[1] = ex.ExampleMetrics(theta_1=10, theta_2=10)
        tr.evaluate_subjects()

        # Only way to move mouse is with override
        tr.override_subject_status(0, override_stage=stageA, override_policies=ex.stageA_policyA)
        tr.override_subject_status(1, override_stage=stageA, override_policies=ex.stageA_policyB)
        tr.override_subject_status(0, override_stage=stageA, override_policies=ex.stageA_policyB)
        tr.override_subject_status(1, override_stage=stageA, override_policies=ex.stageA_policyA)

        trainer = Trainer(curr)
        # Validate mouse history
        M0 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.stageA_policyA]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.stageA_policyB]),
        ]

        M1 = [
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.INIT_STAGE]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.stageA_policyB]),
            trainer.create_trainer_state(stage=stageA, active_policies=[ex.stageA_policyA]),
        ]

        self.assertEqual(tr.subject_history[0], M0)
        self.assertEqual(tr.subject_history[1], M1)

    def test_policy_tracks(self):
        """
        Tests multiple active policies along 2 policy tracks.
        Mouse occupies both tracks.
        """

        curr = ex2.construct_track_curriculum()
        track_stage = curr.see_stages()[0]

        # Create checkpoint stages
        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stage_init = Stage(name="Track Stage", task=dummy_task)
        stage_1 = Stage(name="Track Stage", task=dummy_task)
        stage_2 = Stage(name="Track Stage", task=dummy_task)
        stage_init.set_task_parameters(ex2.DummyParameters(field_1=5, field_2=5))
        stage_1.set_task_parameters(ex2.DummyParameters(field_1=10, field_2=10))
        stage_2.set_task_parameters(ex2.DummyParameters(field_1=15, field_2=15))

        # Associate mice with curriculum
        tr = ex2.ExampleTrainer()
        tr.register_subject(0, curr, track_stage, track_stage.start_policies)

        # Constant mouse metrics
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=10, m2=10)
        tr.evaluate_subjects()
        tr.evaluate_subjects()

        trainer = Trainer(curr)

        M0 = [
            trainer.create_trainer_state(
                stage=stage_init,
                active_policies=[
                    ex2.policy_1,
                    ex2.policy_4,
                ],
            ),
            trainer.create_trainer_state(
                stage=stage_1,
                active_policies=[
                    ex2.policy_2,
                    ex2.policy_5,
                ],
            ),
            trainer.create_trainer_state(
                stage=stage_2,
                active_policies=[
                    ex2.policy_3,
                    ex2.policy_6,
                ],
            ),
        ]

        self.assertEqual(tr.subject_history[0], M0)

    def test_policy_tree(self):
        """
        Tests multiple active policies along policy graph
        structured like a tree.
        Policies that transition into a common policy should
        only be logged once in the SubjectHistory.
        """

        curr = ex2.construct_tree_curriculum()
        tree_stage = curr.see_stages()[0]

        # Create checkpoint stages
        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stage_init = Stage(name="Tree Stage", task=dummy_task)
        stage_1 = Stage(name="Tree Stage", task=dummy_task)
        stage_2 = Stage(name="Tree Stage", task=dummy_task)
        stage_init.set_task_parameters(ex2.DummyParameters(field_1=15, field_2=0))
        stage_1.set_task_parameters(ex2.DummyParameters(field_1=15, field_2=10))
        stage_2.set_task_parameters(ex2.DummyParameters(field_1=15, field_2=15))

        # Associate mice with curriculum
        tr = ex2.ExampleTrainer()
        tr.register_subject(0, curr, tree_stage, tree_stage.start_policies)

        # Constant mouse metrics
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=10, m2=0)
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=10, m2=10)
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        M0 = [
            trainer.create_trainer_state(
                stage=stage_init,
                active_policies=[
                    ex2.policy_1,
                    ex2.policy_2,
                    ex2.policy_3,
                ],
            ),
            trainer.create_trainer_state(
                stage=stage_1,
                active_policies=[
                    ex2.policy_4,
                    ex2.policy_5,
                ],
            ),
            trainer.create_trainer_state(
                stage=stage_1,
                active_policies=[
                    ex2.policy_4,
                    ex2.policy_5,
                ],
            ),
            trainer.create_trainer_state(stage=stage_2, active_policies=[ex2.policy_6]),
        ]

        self.assertEqual(tr.subject_history[0], M0)

    def test_policy_complete_graph(self):
        """
        Tests policy traversal through a complete graph of 3 policies.
        Policies traverse loop 'clockwise', then 'counterclockwise'.
        """

        curr = ex2.construct_policy_triangle_curriculum()
        triangle_stage = curr.see_stages()[0]

        # Create checkpoint stages
        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stage_init = Stage(name="Triangle Stage", task=dummy_task)
        stage_1 = Stage(name="Triangle Stage", task=dummy_task)
        stage_2 = Stage(name="Triangle Stage", task=dummy_task)
        stage_3 = Stage(name="Triangle Stage", task=dummy_task)
        stage_4 = Stage(name="Triangle Stage", task=dummy_task)
        stage_5 = Stage(name="Triangle Stage", task=dummy_task)
        stage_6 = Stage(name="Triangle Stage", task=dummy_task)
        stage_init.set_task_parameters(ex2.DummyParameters(field_1=5))
        stage_1.set_task_parameters(ex2.DummyParameters(field_1=10))
        stage_2.set_task_parameters(ex2.DummyParameters(field_1=15))
        stage_3.set_task_parameters(ex2.DummyParameters(field_1=20))
        stage_4.set_task_parameters(ex2.DummyParameters(field_1=25))
        stage_5.set_task_parameters(ex2.DummyParameters(field_1=30))
        stage_6.set_task_parameters(ex2.DummyParameters(field_1=35))

        # Associate mice with curriculum
        tr = ex2.ExampleTrainer()
        tr.register_subject(0, curr, triangle_stage, triangle_stage.start_policies)

        # Constant mouse metrics
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=10, m2=0)
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=0, m2=0)
        tr.evaluate_subjects()
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=0, m2=10)
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        M0 = [
            trainer.create_trainer_state(stage=stage_init, active_policies=[ex2.policy_1]),
            trainer.create_trainer_state(stage=stage_1, active_policies=[ex2.policy_2]),
            trainer.create_trainer_state(stage=stage_2, active_policies=[ex2.policy_3]),
            # Return to start
            trainer.create_trainer_state(stage=stage_3, active_policies=[ex2.policy_1]),
            trainer.create_trainer_state(stage=stage_3, active_policies=[ex2.policy_1]),
            trainer.create_trainer_state(stage=stage_4, active_policies=[ex2.policy_3]),
            trainer.create_trainer_state(stage=stage_5, active_policies=[ex2.policy_2]),
            trainer.create_trainer_state(stage=stage_6, active_policies=[ex2.policy_1]),
        ]

        self.assertEqual(tr.subject_history[0], M0)

    def test_stage_complete_graph(self):
        """
        Tests stage traversal through a complete graph of 3 stages.
        Subject traverses loop 'clockwise', then 'counterclockwise'.
        """

        curr = ex2.construct_stage_triangle_curriculum()

        stage_1, stage_2, stage_3 = curr.see_stages()
        tr = ex2.ExampleTrainer()
        tr.register_subject(0, curr, stage_1, stage_1.start_policies)

        # Constant mouse metrics
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=10, m2=0)
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=0, m2=0)
        tr.evaluate_subjects()
        ex2.MICE_METRICS[0] = ex2.ExampleMetrics2(m1=0, m2=10)
        tr.evaluate_subjects()
        tr.evaluate_subjects()
        tr.evaluate_subjects()

        trainer = Trainer(curr)
        M0 = [
            trainer.create_trainer_state(stage=stage_1, active_policies=[]),
            trainer.create_trainer_state(stage=stage_2, active_policies=[]),
            trainer.create_trainer_state(stage=stage_3, active_policies=[]),
            trainer.create_trainer_state(stage=stage_1, active_policies=[]),  # Return to start
            trainer.create_trainer_state(stage=stage_1, active_policies=[]),
            trainer.create_trainer_state(stage=stage_3, active_policies=[]),
            trainer.create_trainer_state(stage=stage_2, active_policies=[]),
            trainer.create_trainer_state(stage=stage_1, active_policies=[]),
        ]

        self.assertEqual(tr.subject_history[0], M0)

    def test_eject_subject(self):
        """
        Tests if subject can be ejected and
        re-enrolled into curriculum with override.
        """
        curr = ex.construct_curriculum()

        # See stages is ordered arbitrarily
        # StageA and StageB picked out from observation
        stageA = curr.see_stages()[0]
        stageB = curr.see_stages()[2]

        stageAA = stageA.model_copy(deep=True)
        stageAA.set_task(ex.stageA_policyA_rule(ex.ExampleMetrics(), ex.TaskA(task_parameters=ex.TaskAParameters())))

        tr = ex.ExampleTrainer()
        tr.register_subject(0, curr, stageA, ex.INIT_STAGE)

        ex.MICE_METRICS[0] = ex.ExampleMetrics(theta_1=8)
        tr.evaluate_subjects()
        tr.eject_subject(0)
        tr.evaluate_subjects()
        tr.override_subject_status(0, override_stage=stageB, override_policies=stageB.start_policies)
        trainer = Trainer(curr)
        # Validate mouse history
        M0 = [
            trainer.create_trainer_state(
                stage=stageA,
                is_on_curriculum=True,
                active_policies=[ex.INIT_STAGE],
            ),
            trainer.create_trainer_state(
                stage=stageAA,
                is_on_curriculum=True,
                active_policies=[ex.stageA_policyA],
            ),
            trainer.create_trainer_state(stage=None, is_on_curriculum=False, active_policies=None),
            trainer.create_trainer_state(stage=None, is_on_curriculum=False, active_policies=None),
            trainer.create_trainer_state(stage=stageB, active_policies=[ex.INIT_STAGE]),
        ]
        self.assertEqual(tr.subject_history[0], M0)

    def test_round_trip_trainer_state(self):
        curr = ex.construct_curriculum()
        stageA = curr.see_stages()[0]

        ts = Trainer(curr).create_trainer_state(stage=stageA, active_policies=[ex.stageA_policyA])

        # Serialize from Child
        instance_json = ts.model_dump_json()
        # Deserialize from Child
        recovered = TrainerState.model_validate_json(instance_json)
        self.assertEqual(ts, recovered)


class TrainerStateTests(unittest.TestCase):
    def setUp(self):
        self.curr = ex.construct_curriculum()
        stageA = self.curr.see_stages()[0]
        active_policies = [ex.INIT_STAGE]

        self.state_from_trainer_state = TrainerState(
            curriculum=self.curr,
            stage=stageA,
            active_policies=active_policies,
            is_on_curriculum=True,
        )

        self.trainer = new_trainer = Trainer(curriculum=self.curr)
        self.state_from_trainer = new_trainer.create_trainer_state(
            stage=stageA,
            active_policies=active_policies,
            is_on_curriculum=True,
        )

    def test_trainer_state_is_equal(self):
        # This tests is a bit brittle as it relies on an implementation detail
        # of __eq__ of Stage I also dont think its super useful, so feel free
        # to remove it in the future if it causes problems
        self.assertEqual(self.state_from_trainer_state, self.state_from_trainer)

    def test_trainer_state_dict_is_equal(self):
        dump_from_trainer_state = self.state_from_trainer_state.model_dump()
        dump_from_trainer = self.state_from_trainer.model_dump()
        self.assertEqual(dump_from_trainer, dump_from_trainer_state)
        self.assertEqual(
            self.state_from_trainer_state.model_validate(dump_from_trainer_state),
            self.state_from_trainer.model_validate(dump_from_trainer_state),
        )
        self.assertEqual(
            self.state_from_trainer_state.model_validate(dump_from_trainer),
            self.state_from_trainer.model_validate(dump_from_trainer),
        )

    def test_trainer_state_json_is_equal(self):
        dump_from_trainer_state = self.state_from_trainer_state.model_dump_json()
        dump_from_trainer = self.state_from_trainer.model_dump_json()
        self.assertEqual(dump_from_trainer, dump_from_trainer_state)
        self.assertEqual(
            self.state_from_trainer_state.model_validate_json(dump_from_trainer_state),
            self.state_from_trainer.model_validate_json(dump_from_trainer_state),
        )
        self.assertEqual(
            self.state_from_trainer_state.model_validate_json(dump_from_trainer),
            self.state_from_trainer.model_validate_json(dump_from_trainer),
        )


if __name__ == "__main__":
    unittest.main()



"""
Useful Placeholders when making Curriculums
"""

import os
import subprocess
from pathlib import Path
from typing import Literal, Optional

from jinja2 import Template
from pydantic import Field

from aind_behavior_curriculum.curriculum import Curriculum, Stage
from aind_behavior_curriculum.task import Task, TaskParameters


class Graduated(Task):
    """
    Utility Final Task.
    """

    name: Literal["Graduated"] = "Graduated"
    task_parameters: TaskParameters = Field(default=TaskParameters(), description="Fill w/ Parameter Defaults")


GRADUATED = Stage(name="GRADUATED", task=Graduated())


def export_diagram(curriculum: Curriculum, path: Optional[os.PathLike] = None) -> str:
    """Renders a curriculum to SVG

    Args:
        curriculum (Curriculum): A curriculum to be rendered
        path (Optional[os.PathLike], optional): An optional path to save the SVG asset to.
        If not path is provided, only the str representation of the SVG will be returned.
        Defaults to None.

    Returns:
        str: A literal string representation of the SVG.
    """

    if path:
        path = Path(path)
        if not path.suffix:
            path = path.with_suffix(".svg")
        elif path.suffix != ".svg":
            raise ValueError("Only svg export is allowed.")

    curriculum.validate_curriculum()

    dot_scripts = [_make_curriculum_dot_script(curriculum)]
    stages = curriculum.see_stages()
    for i, stage in enumerate(stages):
        if not (stage.name == "GRADUATED"):
            dot_scripts.append(_make_stage_dot_script(stages.pop(i)))
    if len(stages) > 0:
        dot_scripts.append(_make_stage_dot_script(stages.pop(-1)))

    final_script = "\n".join(dot_scripts)

    # Run graphviz export
    gvpack_command = ["gvpack", "-u"]
    dot_command = ["dot", "-Tsvg"]
    gvpack_process = subprocess.Popen(
        gvpack_command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )
    dot_process = subprocess.Popen(dot_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True)

    gvpack_output, _ = gvpack_process.communicate(input=final_script.encode())

    dot_process_output, _ = dot_process.communicate(input=gvpack_output.decode())
    if path:
        with open(path, "w", encoding="utf-8") as f:
            f.write(dot_process_output)
    return dot_process_output


def _make_stage_dot_script(s: Stage) -> str:
    """
    Stage to dot script conversion.
    """

    template_string = """
        digraph cluster_{{ stage_id }} {
            labelloc="t";
            label={{ stage_name }};

            // Define nodes with increased font visibility
            node [shape=box, style=filled, fontname=Arial, fontsize=12,
            fillcolor=lightblue, color=black];

            // Define nodes
            {% for n in nodes %}
            {{ n }};
            {% endfor %}

            // Define edges
            {% for edge in edges %}
            {{ edge }};
            {% endfor %}
        }
    """
    template = Template(template_string)
    stage_name = '"' + s.name + '"'

    nodes = []
    for node_id, node in s.graph.nodes.items():
        # Add color to start policies
        if node in s.start_policies:
            node_str = f'{node_id} [label="{node.name}", fillcolor="#FFEA00"]'
        else:
            node_str = f'{node_id} [label="{node.name}"]'
        nodes.append(node_str)

    edges = []
    for start_id, edge_list in s.graph.graph.items():
        for i, (edge, dest_id) in enumerate(edge_list):
            # Use 1-indexing for labels
            i = i + 1

            # Edges must be StageTransition or PolicyTransition
            edge_str = f'{start_id} -> {dest_id} [label="({i}) {edge.name}", minlen=2]'
            edges.append(edge_str)

    stage_dot_script = template.render(
        stage_id="".join(s.name.split()),
        stage_name=stage_name,
        nodes=nodes,
        edges=edges,
    )

    return stage_dot_script


def _make_curriculum_dot_script(c: Curriculum) -> str:
    """
    Curriculum to dot script conversion.
    """

    curr_dot_script = """
        digraph cluster_curriculum {
            color="white";
            label={{ curr_name }};
            fontsize=24;

            node [shape=box, style=filled];
            {% for n in nodes %}
            {{ n }}
            {% endfor %}

            {% for edge in edges %}
            {{ edge }};
            {% endfor %}
        }
    """
    template = Template(curr_dot_script)

    # Add curriculum nodes
    nodes = [f'{node_id} [label="{node.name}"]' for node_id, node in c.graph.nodes.items()]

    # Add curriculum edges
    edges = []
    for start_id, edge_list in c.graph.graph.items():
        for i, (edge, dest_id) in enumerate(edge_list):
            # Use 1-indexing for labels
            i = i + 1

            # Edges must be StageTransition or PolicyTransition
            edge_str = f'{start_id} -> {dest_id} [label="({i}) {edge.name}", minlen=2]'
            edges.append(edge_str)

    curriculum_dot_script = template.render(curr_name='"' + c.name + '"', nodes=nodes, edges=edges)

    return curriculum_dot_script


def export_json(curriculum: Curriculum, path: os.PathLike) -> None:
    """
    Export curriculum json to export path
    """

    path = Path(path)
    if not path.suffix == ".json":
        raise ValueError("Please add .json extension to end of json_path.")

    curriculum.validate_curriculum()

    with open(path, "w", encoding="utf-8") as f:
        print(path)
        f.write(curriculum.model_dump_json(indent=3))



"""
Example of Task creation
"""

import json
from typing import Literal

from pydantic import Field, ValidationInfo, field_validator

from aind_behavior_curriculum import Task, TaskParameters


class ExampleTaskParameters(TaskParameters):
    """
    Example Task Parameters
    """

    # Required: Define type annotations for strict type checks.
    # Make fields immutable with Literal type.
    field_1: int = Field(default=0, ge=0.0)
    field_2: int = Field(default=0, ge=0.0)
    field_3: float = Field(default=0.5, ge=0.0, le=1.0)
    field_4: float = Field(default=0.5, ge=0.0, le=1.0)
    field_5: Literal["Immutable Field"] = "Immutable Field"

    # Optional: Add additional validation to fields.
    @field_validator("field_1", "field_2")
    @classmethod
    def check_something(cls, v: int, info: ValidationInfo):
        """Your validation code here"""
        return v


class ExampleTask(Task):
    """
    Example Task
    """

    name: Literal["TaskName"] = "TaskName"
    description: str = Field(default="Ex description of task")

    task_parameters: ExampleTaskParameters = Field(..., description=ExampleTaskParameters.__doc__.strip())


if __name__ == "__main__":
    # Create task, optionally add parameters
    ex_parameters = ExampleTaskParameters(field_2=50, field_4=0.8)
    ex_task = ExampleTask(task_parameters=ex_parameters)
    print(ex_task)

    # Update Task parameters individually
    ex_task.task_parameters.field_1 = 100
    ex_task.task_parameters.field_2 = 200
    print(ex_task)

    # Export/Serialize Task Schema:
    with open("examples/example_project/jsons/task_schema.json", "w") as f:
        json_dict = ExampleTask.model_json_schema()
        json_string = json.dumps(json_dict, indent=4)
        f.write(json_string)

    # Export/Serialize Instance:
    with open("examples/example_project/jsons/task_instance.json", "w") as f:
        json_dict = ex_task.model_dump()
        json_string = json.dumps(json_dict, indent=4)
        f.write(json_string)

    # Import/De-serialize Instance:
    with open("examples/example_project/jsons/task_instance.json", "r") as f:
        json_data = f.read()
    task_instance = ExampleTask.model_validate_json(json_data)
    print(task_instance)



"""
Curriculum Test Suite
"""

import unittest

import example_project as ex
import example_project_2 as ex2
from pydantic import BaseModel, PydanticUserError

from aind_behavior_curriculum import (
    Curriculum,
    Metrics,
    Policy,
    Stage,
    Task,
    create_curriculum,
)
from aind_behavior_curriculum.curriculum import make_task_discriminator


def init_stage_rule(metrics: Metrics, task: Task) -> Task:
    """
    Trivially pass the default
    """
    return task


INIT_STAGE = Policy(init_stage_rule)


class CurriculumTests(unittest.TestCase):
    """Unit tests for Stage/Curriculum De/Serialization"""

    def test_round_trip_without_policies(self):
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        stageA = Stage(name="StageA", task=taskA)

        # Check if the jsons produced by cross serialization are equal
        instance_json = stageA.model_dump_json()
        # Use the generic deserializer
        recovered = Stage.model_validate_json(instance_json)
        recovered_json = recovered.model_dump_json()
        # Compare the two jsons
        self.assertEqual(instance_json, recovered_json)
        self.assertEqual(stageA, recovered)

    def test_round_trip_with_policies(self):
        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        stageA = Stage(name="StageA", task=taskA)

        stageA.add_policy_transition(INIT_STAGE, ex.stageA_policyB, ex.t1_10)
        stageA.add_policy_transition(INIT_STAGE, ex.stageA_policyA, ex.t1_5)
        stageA.add_policy_transition(ex.stageA_policyA, ex.stageA_policyB, ex.t1_10)

        # Serialize from Child
        instance_json = stageA.model_dump_json()
        # Deserialize from Child
        recovered = Stage.model_validate_json(instance_json)
        recovered_json = recovered.model_dump_json()
        self.assertEqual(instance_json, recovered_json)
        self.assertEqual(stageA.see_policies(), recovered.see_policies())
        self.assertEqual(stageA, recovered)

    def test_round_trip_empty_curriculum(self):
        ex_curr = ex.MyCurriculum(name="My Curriculum")

        # Serialize from Child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from Child
        recovered = ex.MyCurriculum.model_validate_json(instance_json)
        self.assertEqual(ex_curr, recovered)

        # Serialize from Child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from Parent
        instance_parent = Curriculum.model_validate_json(instance_json)
        # Serialize from Parent
        parent_json = instance_parent.model_dump_json()
        # Deserialize from Child
        instance_prime = ex.MyCurriculum.model_validate_json(parent_json)
        self.assertEqual(ex_curr, instance_prime)

    def test_round_trip_curriculum(self):
        ex_curr = ex.construct_curriculum()

        # Serialize from Child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from Child
        recovered = ex.MyCurriculum.model_validate_json(instance_json)

        self.assertEqual(ex_curr, recovered)

        # Serialize from Child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from Parent
        instance_parent = Curriculum.model_validate_json(instance_json)
        # Serialize from Parent
        parent_json = instance_parent.model_dump_json()
        # Deserialize from Child
        instance_prime = ex.MyCurriculum.model_validate_json(parent_json)
        self.assertEqual(ex_curr, instance_prime)

    def test_round_trip_edit_task_parameters(self):
        ex_curr = ex.construct_curriculum()

        stage_0 = ex_curr.see_stages()[0]
        params = stage_0.get_task_parameters()
        params.field_a = 8
        stage_0.set_task_parameters(params)

        # Serialize from child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from child
        recovered = ex.MyCurriculum.model_validate_json(instance_json)
        self.assertEqual(ex_curr, recovered)

        # Serialize from Child
        instance_json = ex_curr.model_dump_json()
        # Deserialize from Parent
        instance_parent = Curriculum.model_validate_json(instance_json)
        # Serialize from Parent
        parent_json = instance_parent.model_dump_json()
        # Deserialize from Child
        instance_prime = ex.MyCurriculum.model_validate_json(parent_json)
        self.assertEqual(ex_curr, instance_prime)
        self.assertEqual(instance_prime.see_stages()[0], stage_0)

    def test_add_policies_and_policy_transitions(self):
        """
        Constructing a diamond graph and checking if
        stored polices and policy transitions are correct.

        For diamond graph:
        A -> B -> D
        A -> C -> D

        Testing:
        Stage contains [A, B, C, D]
        A's transitions are [B, C]
        B's transitions are [D]
        C's transitions are [D]
        D's transitions are [].
        """

        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)

        stageA.add_policy(ex2.policy_1)
        stageA.set_start_policies(ex2.policy_1)

        stageA.add_policy_transition(ex2.policy_1, ex2.policy_2, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_1, ex2.policy_3, ex2.m1_policy_transition)

        stageA.add_policy_transition(ex2.policy_2, ex2.policy_4, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_3, ex2.policy_4, ex2.m1_policy_transition)

        self.assertEqual(
            stageA.see_policies(),
            [ex2.policy_1, ex2.policy_2, ex2.policy_3, ex2.policy_4],
        )
        self.assertEqual(
            stageA.see_policy_transitions(ex2.policy_1),
            [
                (ex2.m1_policy_transition, ex2.policy_2),
                (ex2.m1_policy_transition, ex2.policy_3),
            ],
        )
        self.assertEqual(
            stageA.see_policy_transitions(ex2.policy_2),
            [(ex2.m1_policy_transition, ex2.policy_4)],
        )
        self.assertEqual(
            stageA.see_policy_transitions(ex2.policy_3),
            [(ex2.m1_policy_transition, ex2.policy_4)],
        )
        self.assertEqual(stageA.see_policy_transitions(ex2.policy_4), [])

    def test_add_stage_and_stage_transitions(self):
        """
        Constructing a diamond graph and checking if
        stored stages and stage transitions are correct.

        For diamond graph:
        A -> B -> D
        A -> C -> D

        Testing:
        Curr contains [A, B, C, D]
        A's transitions are [B, C]
        B's transitions are [D]
        C's transitions are [D]
        D's transitions are [].
        """

        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)
        stageB = Stage(name="Stage B", task=dummy_task)
        stageC = Stage(name="Stage C", task=dummy_task)
        stageD = Stage(name="Stage D", task=dummy_task)

        ex_curr = ex2.MyCurriculum()
        ex_curr.add_stage(stageA)
        ex_curr.add_stage_transition(stageA, stageB, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageA, stageC, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageB, stageD, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageC, stageD, ex2.m1_stage_transition)

        self.assertEqual(ex_curr.see_stages(), [stageA, stageB, stageC, stageD])
        self.assertEqual(
            ex_curr.see_stage_transitions(stageA),
            [
                (ex2.m1_stage_transition, stageB),
                (ex2.m1_stage_transition, stageC),
            ],
        )
        self.assertEqual(
            ex_curr.see_stage_transitions(stageB),
            [(ex2.m1_stage_transition, stageD)],
        )
        self.assertEqual(
            ex_curr.see_stage_transitions(stageC),
            [(ex2.m1_stage_transition, stageD)],
        )
        self.assertEqual(ex_curr.see_stage_transitions(stageD), [])

    def test_reorder_policy_transitions(self):
        """
        Constructing a diamond graph and reordering
        priority of two branches.

        For diamond graph:
        A -> B -> D
        A -> C -> D
        and swapping priority of B and C.

        Testing:
        A's transitions are reordered
        from [B, C] to [C, B].
        """
        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)

        stageA.add_policy(ex2.policy_1)
        stageA.set_start_policies(ex2.policy_1)

        stageA.add_policy_transition(ex2.policy_1, ex2.policy_2, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_1, ex2.policy_3, ex2.m1_policy_transition)

        stageA.add_policy_transition(ex2.policy_2, ex2.policy_4, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_3, ex2.policy_4, ex2.m1_policy_transition)

        new_priority = [
            (ex2.m1_policy_transition, ex2.policy_3),
            (ex2.m1_policy_transition, ex2.policy_2),
        ]
        stageA.set_policy_transition_priority(ex2.policy_1, new_priority)

        self.assertEqual(stageA.see_policy_transitions(ex2.policy_1), new_priority)

    def test_remove_policies_and_policy_transitions(self):
        """
        Tests if policies can be removed in all
        the ways offered.

        For diamond graph:
        A -> B -> D
        A -> C -> D
        Testing:
        - Removing A
        - Removing D
        - Removing B
        - Removing all edges and creating floating policies.
        """

        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)

        stageA.add_policy(ex2.policy_1)
        stageA.set_start_policies(ex2.policy_1)

        stageA.add_policy_transition(ex2.policy_1, ex2.policy_2, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_1, ex2.policy_3, ex2.m1_policy_transition)

        stageA.add_policy_transition(ex2.policy_2, ex2.policy_4, ex2.m1_policy_transition)
        stageA.add_policy_transition(ex2.policy_3, ex2.policy_4, ex2.m1_policy_transition)

        stageA1 = stageA.model_copy(deep=True)
        stageA2 = stageA.model_copy(deep=True)
        stageA3 = stageA.model_copy(deep=True)
        stageA4 = stageA.model_copy(deep=True)

        stageA1.remove_policy(ex2.policy_1)
        self.assertEqual(stageA1.see_policies(), [ex2.policy_2, ex2.policy_3, ex2.policy_4])

        stageA2.remove_policy(ex2.policy_4)
        self.assertEqual(stageA2.see_policy_transitions(ex2.policy_2), [])
        self.assertEqual(stageA2.see_policy_transitions(ex2.policy_3), [])

        stageA3.remove_policy(ex2.policy_2)
        self.assertEqual(
            stageA3.see_policy_transitions(ex2.policy_1),
            [(ex2.m1_policy_transition, ex2.policy_3)],
        )

        stageA4.remove_policy_transition(ex2.policy_1, ex2.policy_2, ex2.m1_policy_transition)
        stageA4.remove_policy_transition(ex2.policy_1, ex2.policy_3, ex2.m1_policy_transition)
        stageA4.remove_policy_transition(ex2.policy_2, ex2.policy_4, ex2.m1_policy_transition)
        stageA4.remove_policy_transition(ex2.policy_3, ex2.policy_4, ex2.m1_policy_transition)
        self.assertEqual(stageA4.see_policy_transitions(ex2.policy_1), [])
        self.assertEqual(stageA4.see_policy_transitions(ex2.policy_2), [])
        self.assertEqual(stageA4.see_policy_transitions(ex2.policy_3), [])
        self.assertEqual(stageA4.see_policy_transitions(ex2.policy_4), [])

    def test_remove_stages_and_stage_transitions(self):
        """
        Tests if stages can be removed in all
        the ways offered.

        For diamond graph:
        A -> B -> D
        A -> C -> D
        Testing:
        - Removing A
        - Removing D
        - Removing B
        - Removing all edges and creating floating stages.
        """

        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)
        stageB = Stage(name="Stage B", task=dummy_task)
        stageC = Stage(name="Stage C", task=dummy_task)
        stageD = Stage(name="Stage D", task=dummy_task)

        ex_curr = ex2.MyCurriculum()
        ex_curr.add_stage(stageA)
        ex_curr.add_stage_transition(stageA, stageB, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageA, stageC, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageB, stageD, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageC, stageD, ex2.m1_stage_transition)

        ex_curr1 = ex_curr.model_copy(deep=True)
        ex_curr2 = ex_curr.model_copy(deep=True)
        ex_curr3 = ex_curr.model_copy(deep=True)
        ex_curr4 = ex_curr.model_copy(deep=True)

        ex_curr1.remove_stage(stageA)
        self.assertEqual(ex_curr1.see_stages(), [stageB, stageC, stageD])

        ex_curr2.remove_stage(stageD)
        self.assertEqual(ex_curr2.see_stage_transitions(stageB), [])
        self.assertEqual(ex_curr2.see_stage_transitions(stageC), [])

        ex_curr3.remove_stage(stageB)
        self.assertEqual(
            ex_curr3.see_stage_transitions(stageA),
            [(ex2.m1_stage_transition, stageC)],
        )

        ex_curr4.remove_stage_transition(stageA, stageB, ex2.m1_stage_transition)
        ex_curr4.remove_stage_transition(stageA, stageC, ex2.m1_stage_transition)
        ex_curr4.remove_stage_transition(stageB, stageD, ex2.m1_stage_transition)
        ex_curr4.remove_stage_transition(stageC, stageD, ex2.m1_stage_transition)
        self.assertEqual(ex_curr4.see_stage_transitions(stageA), [])
        self.assertEqual(ex_curr4.see_stage_transitions(stageB), [])
        self.assertEqual(ex_curr4.see_stage_transitions(stageC), [])
        self.assertEqual(ex_curr4.see_stage_transitions(stageD), [])

    def test_reorder_stage_transitions(self):
        """
        Constructing a diamond graph and reordering
        priority of two branches.

        For diamond graph:
        A -> B -> D
        A -> C -> D
        and swapping priority of B and C.

        Testing:
        A's transitions are reordered
        from [B, C] to [C, B].
        """

        dummy_task = ex2.DummyTask(task_parameters=ex2.DummyParameters())
        stageA = Stage(name="Stage A", task=dummy_task)
        stageB = Stage(name="Stage B", task=dummy_task)
        stageC = Stage(name="Stage C", task=dummy_task)
        stageD = Stage(name="Stage D", task=dummy_task)

        ex_curr = ex2.MyCurriculum()
        ex_curr.add_stage(stageA)
        ex_curr.add_stage_transition(stageA, stageB, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageA, stageC, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageB, stageD, ex2.m1_stage_transition)
        ex_curr.add_stage_transition(stageC, stageD, ex2.m1_stage_transition)

        new_priority = [
            (ex2.m1_stage_transition, stageC),
            (ex2.m1_stage_transition, stageB),
        ]
        ex_curr.set_stage_transition_priority(stageA, new_priority)

        self.assertEqual(ex_curr.see_stage_transitions(stageA), new_priority)

    def test_create_curriculum(self):
        _ = create_curriculum("test_curriculum", "1.2.3", (ex.TaskA, ex.TaskB))
        _ = create_curriculum("test_curriculum", "1.2.3", (ex.TaskA, ex.TaskB, ex.TaskB))
        _ = create_curriculum(
            "test_curriculum",
            "1.2.3",
            (ex.TaskA, ex.TaskB),
            pkg_location="example_project",
        )

    def test_create_curriculum_equivalence(self):
        _tasks = make_task_discriminator((ex.TaskA, ex.TaskB))

        class TestCurriculum(Curriculum[_tasks]):
            name: str = "TestCurriculum"
            version: str = "1.2.3"
            pkg_location: str = "test"

        taskA = ex.TaskA(task_parameters=ex.TaskAParameters())
        taskB = ex.TaskB(task_parameters=ex.TaskBParameters())
        expected_curriculum = TestCurriculum()
        expected_curriculum.add_stage(Stage(name="Stage 0", task=taskA))
        expected_curriculum.add_stage(Stage(name="Stage 1", task=taskB))

        created_curriculum = create_curriculum(
            "TestCurriculum",
            "1.2.3",
            (ex.TaskA, ex.TaskB),
            pkg_location="test",
        )()
        created_curriculum.add_stage(Stage(name="Stage 0", task=taskA))
        created_curriculum.add_stage(Stage(name="Stage 1", task=taskB))

        self.assertEqual(expected_curriculum.model_dump(), created_curriculum.model_dump())
        self.assertEqual(
            expected_curriculum.model_dump_json(),
            created_curriculum.model_dump_json(),
        )
        self.assertEqual(
            expected_curriculum.model_validate_json(expected_curriculum.model_dump_json()),
            expected_curriculum.model_validate_json(expected_curriculum.model_dump_json()),
        )

    def test_create_curriculum_with_invalid_tagged_union(self):
        class NotATask(BaseModel):
            not_name: str = "Not a Task"

        with self.assertRaises(PydanticUserError) as _:
            _ = create_curriculum("test_curriculum", "1.2.3", (ex.TaskA, ex.TaskB, NotATask))


if __name__ == "__main__":
    unittest.main()
